/*******************************************************************************

  Memory IO
 
  Programmer: Dima Fedorov Levit <dimin@dimin.net> <http://www.dimin.net/>
  
  Based on memio used by GDAL and libtiff by:
      Mike Johnson - Banctec AB
      Frank Warmerdam, warmerdam@pobox.com

  History:
    08/05/2004 13:57 - First creation
      
  ver: 1
        
*******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "dmemio.h"

static void dimMemIO_ExtendFile( TDimMemIO *miobuf, ulong size );

//------------------------------------------------------------------------
// Initialize a passed in TDimMemIO structure.                      
//------------------------------------------------------------------------

void  dimMemIO_Init( TDimMemIO *miobuf, int size, unsigned char *data )
{
  miobuf->data = NULL;
  miobuf->size = 0;
  miobuf->offset = 0;
  miobuf->data_buf_size = 0;
  miobuf->own_buffer = 1;

  if( size > 0 )
  {
      miobuf->data = data;
      miobuf->size = size;
      miobuf->data_buf_size = size;
      miobuf->own_buffer = 0;
  }
}

//------------------------------------------------------------------------
// Clear and free memory buffer
//------------------------------------------------------------------------

void  dimMemIO_Destroy( TDimMemIO *miobuf )
{
  if( miobuf->own_buffer && miobuf->data != NULL )
      free( miobuf->data );
  memset( miobuf, 0, sizeof(TDimMemIO) );
}

//------------------------------------------------------------------------
// seek within the buffer
//------------------------------------------------------------------------

int dimMemIO_Seek ( void *stream, long offset, int origin )
{
  TDimMemIO *miobuf = (TDimMemIO *) stream;
  long new_off;

  if( origin == SEEK_SET )
    new_off = offset;
  else if( origin == SEEK_CUR )
    new_off = miobuf->offset + offset;
  else if( origin == SEEK_END )
    new_off = miobuf->size + offset;
  else
    return -1;

  if( new_off < 0 ) return -1;

  if( new_off > (long) miobuf->size )
  {
    dimMemIO_ExtendFile( miobuf, new_off );
    if ( new_off > (long) miobuf->size ) return -1;
  }
  
  miobuf->offset = new_off;
  
  return miobuf->offset;
}

//------------------------------------------------------------------------
// read from the buffer
//------------------------------------------------------------------------

ulong dimMemIO_Read ( void *buffer, ulong size, ulong count, void *stream )
{
  TDimMemIO *miobuf = (TDimMemIO *) stream;
  int result = 0;
  ulong nsize = size * count;

  if( miobuf->offset + nsize > miobuf->size )
    result = miobuf->size - miobuf->offset;
  else
    result = nsize;

  memcpy( buffer, miobuf->data + miobuf->offset, result );
  miobuf->offset += result;

  return (ulong) floor( (double) result / (double) size );
}

//------------------------------------------------------------------------
// write into buffer
//------------------------------------------------------------------------

ulong dimMemIO_Write ( void *buffer, ulong size, ulong count, void *stream )
{
  TDimMemIO *miobuf = (TDimMemIO *) stream;
  int result = 0;
  ulong nsize = size * count;

  if( miobuf->offset + nsize > miobuf->size )
      dimMemIO_ExtendFile( miobuf, miobuf->offset + nsize );

  if( miobuf->offset + nsize > miobuf->size )
    result = miobuf->size - miobuf->offset;
  else
    result = nsize;

  memcpy( miobuf->data + miobuf->offset, buffer, result );
  miobuf->offset += result;

  return (ulong) floor( (double)result / (double)size );
}

//------------------------------------------------------------------------
// access buffer size
//------------------------------------------------------------------------

ulong dimMemIO_Size ( void *stream )
{
  TDimMemIO *miobuf = (TDimMemIO *) stream;
  return miobuf->size;
}

//------------------------------------------------------------------------
// flush the buffer - no such thing
//------------------------------------------------------------------------

int dimMemIO_Flush ( void *stream )
{
  return 0;
}

//------------------------------------------------------------------------
// return current position
//------------------------------------------------------------------------

long dimMemIO_Tell ( void *stream )
{
  TDimMemIO *miobuf = (TDimMemIO *) stream;
  return miobuf->offset;
}

//------------------------------------------------------------------------
// test if current position is in the end of the file
//------------------------------------------------------------------------

int dimMemIO_Eof ( void *stream )
{
  TDimMemIO *miobuf = (TDimMemIO *) stream;

  if ( miobuf->offset == miobuf->size ) 
    return 1;
  else
    return 0;
}

//------------------------------------------------------------------------
// prototype for close stream funct, does nothing
//------------------------------------------------------------------------

int dimMemIO_Close ( void *stream )
{
  return 0;
}


//------------------------------------------------------------------------
// increase buffer size
//------------------------------------------------------------------------

static void dimMemIO_ExtendFile( TDimMemIO *miobuf, ulong size )
{
  ulong new_buf_size;
  uchar *new_buffer;

  if( size < miobuf->size ) return;

  if( size < miobuf->data_buf_size ) {
      miobuf->size = size;
      return;
  }

  new_buf_size = (int) (size * 1.25);
  
  if( miobuf->own_buffer )
    new_buffer = (uchar *) realloc( miobuf->data, new_buf_size );
  else
  {
    new_buffer = (uchar *) malloc( new_buf_size );
    if( new_buffer != NULL )
      memcpy( new_buffer, miobuf->data, miobuf->size );
  }
  
  if( new_buffer == NULL ) return;

  miobuf->data = new_buffer;
  miobuf->data_buf_size = new_buf_size;
  miobuf->size = size;
}

