/*******************************************************************************

  Manager for Image Formats with MetaData parsing

  Uses DimFiSDK version: 1
  
  Programmer: Dima Fedorov Levit <dimin@dimin.net> <http://www.dimin.net/>

  Notes:
    Session: during session any session wide operation will be performed
    with current session!!!
    If you want to start simultaneous sessions then create new manager
    using = operator. It will copy all necessary initialization data
    to the new manager.

    Non-session wide operation might be performed simultaneously with 
    an open session.

  History:
    03/23/2004 18:03 - First creation
    01/25/2007 21:00 - added QImaging TIFF
      
  ver: 3
        

*******************************************************************************/

#include <string>
#include <sstream>
#include <xstring.h>
#include <map>
#include <vector>

#include "meta_format_manager.h"
#include "tiff/dim_tiff_format.h"

// windows: use secure C libraries with VS2005 or higher
#if ( defined(_MSC_VER) && (_MSC_VER >= 1400) )
  #pragma warning(disable:4996)
  #define HAVE_SECURE_C
  #pragma message(">>>>> meta_format_manager: using secure c libraries")
#endif 

#ifdef HAVE_SECURE_C  
  #define SSCANF sscanf_s
  #define SPRINTF sprintf_s
#else
  #define SSCANF sscanf
  #define SPRINTF sprintf
#endif

#ifndef min
#define min(a, b)   (((a) < (b)) ? (a) : (b)) 
#endif

const char months[13][4] = {"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };

int getMonthNum(const char *buf) {
  for (int i=1; i<=12; i++)
    if (strncmp( months[i], buf, 3 ) == 0) return i;
  return 0;
}

float scan_float( const char* str, const char* title, float def = 0.0 ) {
  float r = def;
  char *line = (char*) strstr( str, title );
  std::string fmt_str = title;
  fmt_str += "%f";
  if (line != NULL)
    SSCANF( line, fmt_str.c_str(), &r );
  return r;
}

std::string scan_string( const char* str, const char* title ) {
  std::string r;
  char buf[4048];
  buf[0] = 0;

  char *line = (char*) strstr( str, title );
  if (line != NULL) {
    std::string fmt_str = title;
    fmt_str += "%s\n";
    try {
      if (sscanf( line, fmt_str.c_str(), buf ))
        r = buf;
    } catch (...) { }
  }
  return r;
}

TMetaFormatManager::TMetaFormatManager() 
: TDimFormatManager() 
{

}

TMetaFormatManager::~TMetaFormatManager()
{

}

int TMetaFormatManager::sessionStartRead(const char *fileName)
{
  got_meta_for_session = -1;
#ifdef HAVE_SECURE_C  
  sprintf_s(imaging_time, "0000-00-00 00:00:00", 19);
#else
  sprintf(imaging_time, "0000-00-00 00:00:00");
#endif

  pixel_size[0] = 0;
  pixel_size[1] = 0;
  pixel_size[2] = 0;
  pixel_size[3] = 0;
  channel_names.clear();
  display_lut.clear();
  metadata.clear();
  return TDimFormatManager::sessionStartRead(fileName);
}

void TMetaFormatManager::sessionEnd()
{
  got_meta_for_session = -1;
  TDimFormatManager::sessionEnd();
}

int TMetaFormatManager::sessionReadImage ( TDimImageBitmap *bmp, DIM_UINT page ) {
  int res = TDimFormatManager::sessionReadImage( bmp, page );
  
  channel_names.clear();
  display_lut.clear();
  metadata.clear();

  display_lut.resize(3);
  if (bmp->i.samples == 1)
    for (int i=0; i<3; ++i) display_lut[i] = 0;
  else
    if (bmp->i.samples == 2) {
      for (int i=0; i<2; ++i) display_lut[i] = i;
      display_lut[2] = -1;
    } 
  else
  if (bmp->i.samples >= 3)
    for (int i=0; i<3; ++i) display_lut[i] = i;

  switch ( bmp->i.resUnits )
  {
    case DIM_RES_um:
      pixel_size[0] = bmp->i.xRes;
      pixel_size[1] = bmp->i.xRes;
      pixel_size[2] = 0;
      pixel_size[3] = 0;
      break;

    case DIM_RES_nm:
      pixel_size[0] = bmp->i.xRes / 1000.0;
      pixel_size[1] = bmp->i.xRes / 1000.0;
      pixel_size[2] = 0;
      pixel_size[3] = 0;
      break;

    case DIM_RES_mm:
      pixel_size[0] = bmp->i.xRes * 1000.0;
      pixel_size[1] = bmp->i.xRes * 1000.0;
      pixel_size[2] = 0;
      pixel_size[3] = 0;
      break;

    default:
      pixel_size[0] = 0;
      pixel_size[1] = 0;
      pixel_size[2] = 0;
      pixel_size[3] = 0;
  }

  return res;
}

void TMetaFormatManager::sessionParseMetaData ( DIM_UINT page )
{
  tagList = sessionReadMetaData(page, -1, -1, -1);
  got_meta_for_session = page;

  char *meta_char = sessionGetTextMetaData();
  if (meta_char != NULL) {
    meta_data_text = meta_char;
    delete meta_char;
  }
  
  parse_meta_data();

  appendMetadata( "format", this->sessionGetFormatName() );
  if ( strncmp( this->getImagingTime(), "0000-00-00 00:00:00", 19 ) != 0 )   
    appendMetadata( "date_time", this->getImagingTime() );
  
  if ( this->getPixelSizeX() > 0 ) {
    appendMetadata( "pixel_resolution_x", this->getPixelSizeX() );
    appendMetadata( "pixel_resolution_unit_x", "microns" );
  }

  if ( this->getPixelSizeY() > 0 ) {
    appendMetadata( "pixel_resolution_y", this->getPixelSizeY() );
    appendMetadata( "pixel_resolution_unit_y", "microns" );
  }

  if ( this->getPixelSizeZ() > 0 ) {
    appendMetadata( "pixel_resolution_z", this->getPixelSizeZ() );
    appendMetadata( "pixel_resolution_unit_z", "microns" );
  }

  if ( this->getPixelSizeT() > 0 ) {
    appendMetadata( "pixel_resolution_t", this->getPixelSizeT() );
    appendMetadata( "pixel_resolution_unit_t", "seconds" );
  }

  // Image info
  TDimImageBitmap *ibm = sessionHandle.image;
  if (!ibm || !ibm->bits[0]) return;

  appendMetadata( "image_num_x", (int) ibm->i.width );
  appendMetadata( "image_num_y", (int) ibm->i.height );
  appendMetadata( "image_num_z", (int) ibm->i.number_z );
  appendMetadata( "image_num_t", (int) ibm->i.number_t );
  appendMetadata( "image_num_c", (int) ibm->i.samples );
  appendMetadata( "image_num_p", (int) ibm->i.number_pages );
  appendMetadata( "image_pixel_depth", (int) ibm->i.depth );

  for (unsigned int i=0; i<channel_names.size(); ++i) {
    xstring tag_name;
    tag_name.sprintf( "channel_%d_name", i );
    appendMetadata( tag_name.c_str(), channel_names[i].c_str() );
  }

  if (display_lut.size() >= 3) {
    appendMetadata( "display_channel_red",   display_lut[0] );
    appendMetadata( "display_channel_green", display_lut[1] );
    appendMetadata( "display_channel_blue",  display_lut[2] );
  }
}

double TMetaFormatManager::getPixelSizeX()
{
  if (got_meta_for_session != sessionGetCurrentPage() ) 
    sessionParseMetaData ( sessionGetCurrentPage() );
  return pixel_size[0];
}

double TMetaFormatManager::getPixelSizeY()
{
  if (got_meta_for_session != sessionGetCurrentPage() ) 
    sessionParseMetaData ( sessionGetCurrentPage() );
  return pixel_size[1];
}

double TMetaFormatManager::getPixelSizeZ()
{
  if (got_meta_for_session != sessionGetCurrentPage() ) 
    sessionParseMetaData ( sessionGetCurrentPage() );
  return pixel_size[2];
}

double TMetaFormatManager::getPixelSizeT()
{
  if (got_meta_for_session != sessionGetCurrentPage() ) 
    sessionParseMetaData ( sessionGetCurrentPage() );
  return pixel_size[3];
}

const char* TMetaFormatManager::getImagingTime() {
  if (got_meta_for_session != sessionGetCurrentPage() ) 
    sessionParseMetaData ( sessionGetCurrentPage() );
  return imaging_time;
}


TDimImageBitmap *TMetaFormatManager::sessionImage()
{
  //return &bmp;
  return NULL;
}

//***************************************************************************************
// SPECIFIC METADATA PARSING
//***************************************************************************************


//---------------------------------------------------------------------------------------
// FLUOVIEW
//---------------------------------------------------------------------------------------

void TMetaFormatManager::parse_fluoview_285 (char *buf) {
  char *line;
  float xr=0.0, yr=0.0;

  line = strstr( buf, "Resolution" );

  if (line != NULL) {
    SSCANF( line, "Resolution\t%f\t%f", &xr, &yr );
    pixel_size[0] = xr;
    pixel_size[1] = yr;
  }
}

void TMetaFormatManager::parse_fluoview_270 (char *buf) {
  char *line;
  int d=0, m=0, y=0, h=0, mi=0, s=0;
  
  //Date=02-17-2004
  line = strstr( buf, "Date=" );
  if (line != NULL)
    SSCANF( line, "Date=%d-%d-%d", &m, &d, &y );

  //Time=11:54:50
  line = strstr( buf, "Time=" );
  if (line != NULL)
    SSCANF( line, "Time=%d:%d:%d", &h, &mi, &s );

  //md->imaging_time = new char [20];
  SPRINTF(imaging_time, "%.4d-%.2d-%.2d %.2d:%.2d:%.2d", y, m, d, h, mi, s);

}

void TMetaFormatManager::parse_fluoview() {
  if (tagList == NULL) return;
  if (tagList->count == 0) return;
  if (tagList->tags == NULL) return; 
  int pos; 

  if (isTagPresent( tagList, DIM_META_TIFF_TAG, 34361 ) == TRUE) {
    
    if ((pos = tagPos(tagList, DIM_META_TIFF_TAG, 270)) != -1)
      parse_fluoview_270 ( (char *) tagList->tags[pos].tagData );

    if ((pos = tagPos(tagList, DIM_META_TIFF_TAG, 285)) != -1)
      parse_fluoview_285 ( (char *) tagList->tags[pos].tagData );

    // get Z resolution from metadata text
    if (meta_data_text.size() > 0) {
      pixel_size[2] = scan_float( meta_data_text.c_str(), "Z Resolution: " );

      //Magnification=40X
      std::string mag_str = scan_string( meta_data_text.c_str(), "Magnification=" );
      appendMetadata( "magnification", mag_str );
    }

    // now extract channel names
    TDimTiffParams *tiffpar = (TDimTiffParams *) sessionHandle.internalParams;
    if (tiffpar) {
      TDimFluoviewInfo *fvi = &tiffpar->fluoviewInfo;
      if ( fvi->sample_names.size() > 0 )
        channel_names.assign( fvi->sample_names.begin(), fvi->sample_names.end() );

      if ( fvi->display_lut.size() == 3 )
        display_lut.assign( fvi->display_lut.begin(), fvi->display_lut.end() );
    }

  } // if DIM_META_TIFF_TAG 34361
}

//---------------------------------------------------------------------------------------
// Carl Zeiss LSM 5
//---------------------------------------------------------------------------------------

void TMetaFormatManager::parse_lsm() {
  // let's simply rely on info pre parsed in the format and get it from text
  if (meta_data_text.size() <= 0) return;

  pixel_size[0] = scan_float( meta_data_text.c_str(), "X Resolution: " );
  pixel_size[1] = scan_float( meta_data_text.c_str(), "Y Resolution: " );
  pixel_size[2] = scan_float( meta_data_text.c_str(), "Z Resolution: " );
  pixel_size[3] = scan_float( meta_data_text.c_str(), "T Resolution: " );
}

//---------------------------------------------------------------------------------------
// BIORAD
//---------------------------------------------------------------------------------------
void TMetaFormatManager::parse_biorad_20 (char *buf)
{
  char *line;
  int d=0;
  float f=0.0, res=0.0;
  char t[1024];
  
  // Common:
  //AXIS_2 = 257 0.000000e+00 2.689307e-01 microns 
  // Uncommon way
  //AXIS_2 257 0.000000e+00 2.286561e-01 microns

  // X
  line = strstr( buf, "AXIS_2 =" );
  if (line != NULL)
    SSCANF( line, "AXIS_2 = %d %e %e microns ", &d, &f, &res );
  else
  {
    line = strstr( buf, "AXIS_2 " );
    if (line != NULL)
      sscanf( line, "%s %d %e %e microns ", t, &d, &f, &res );
  }
  pixel_size[0] = res;    

  // Y
  line = strstr( buf, "AXIS_3 =" );
  if (line != NULL)
    SSCANF( line, "AXIS_3 = %d %e %e microns ", &d, &f, &res );
  else  
  {
    line = strstr( buf, "AXIS_3 " );
    if (line != NULL)
      sscanf( line, "%s %d %e %e microns ", t, &d, &f, &res );
  }
  pixel_size[1] = res;

  // Z
  line = strstr( buf, "AXIS_4 =" );
  if (line != NULL)
    SSCANF( line, "AXIS_4 = %d %e %e microns ", &d, &f, &res );
  else  
  {
    line = strstr( buf, "AXIS_4 " );
    if (line != NULL)
      sscanf( line, "%s %d %e %e microns ", t, &d, &f, &res );
  }
  pixel_size[2] = res;
}

void TMetaFormatManager::parse_biorad_01 (char *buf)
{
  char t[1024];
  char month[3];
  int m=0, y, d, h, mi, s;
  //Live Thu Aug  7 13:42:13 2003  Zoom 2.0 Kalman 3 Mixer A PMT 

  //sscanf( buf, "%s %s %3s %d %d:%d:%d %d", &t, &t, &month, &d, &h, &mi, &s, &y ); 
  sscanf( buf, "Live %s %3s %d %d:%d:%d %d", t, month, &d, &h, &mi, &s, &y ); 
  m = getMonthNum( month );
  SPRINTF(imaging_time, "%.4d-%.2d-%.2d %.2d:%.2d:%.2d", y, m, d, h, mi, s);
}

void TMetaFormatManager::parse_biorad() {
  if (tagList == NULL) return;
  if (tagList->count == 0) return;
  if (tagList->tags == NULL) return; 
  int pos; 

  if (isTagPresent( tagList, DIM_META_BIORAD, 20 ) == TRUE)
  {
    if ((pos = tagPos(tagList, DIM_META_BIORAD, 20)) != -1)
    if (tagList->tags[pos].tagGroup == DIM_META_BIORAD)
      parse_biorad_20 ( (char *) tagList->tags[pos].tagData );
  }

  if (isTagPresent( tagList, DIM_META_BIORAD, 1 ) == TRUE)
  {
    if ((pos = tagPos(tagList, DIM_META_BIORAD, 1)) != -1)
    if (tagList->tags[pos].tagGroup == DIM_META_BIORAD)
      parse_biorad_01 ( (char *) tagList->tags[pos].tagData );
  }

}

//---------------------------------------------------------------------------------------
// STK
//---------------------------------------------------------------------------------------
void TMetaFormatManager::parse_stk_33629 (char *buf)
{
  char *line;
  xstring query;
  int i=0, y=0, m=0, d=0, h=0, mi=0, s=0;

  query.sprintf("Page %.3d: ", sessionGetCurrentPage());
  line = strstr( buf, query.c_str() );
  if (line != NULL)
    SSCANF( line, "Page %d: %d-%d-%d %d:%d:%d", &i, &y, &m, &d, &h, &mi, &s );
  SPRINTF(imaging_time, "%.4d-%.2d-%.2d %.2d:%.2d:%.2d", y, m, d, h, mi, s);
  
  xstring key;
  xstring plane_val;
  for (unsigned int p=0; p<(unsigned int)sessionGetNumberOfPages(); ++p) {
    query.sprintf("Page %.3d: ", p);
    line = strstr( buf, query.c_str() );
    if (line) {
      SSCANF( line, "Page %d: %d-%d-%d %d:%d:%d", &i, &y, &m, &d, &h, &mi, &s );

      key.sprintf( "date_time_plane_%d", p );
      plane_val.sprintf( "%.4d-%.2d-%.2d %.2d:%.2d:%.2d", y, m, d, h, mi, s );
      appendMetadata( key, plane_val );
    } // if line
  } // for pages
}

void TMetaFormatManager::parse_stk() {
  if (tagList == NULL) return;
  if (tagList->count == 0) return;
  if (tagList->tags == NULL) return;  
  int pos;  

  if (isTagPresent( tagList, DIM_META_TIFF_TAG, 33628 ) == TRUE) {
    // read STK tag 306 - Date/Time
    if ((pos = tagPos(tagList, DIM_META_STK, 33629)) != -1)
      parse_stk_33629 ( (char *) tagList->tags[pos].tagData );

    pixel_size[0] = 0;
    pixel_size[1] = 0;
    pixel_size[2] = 0;
  }
}

//---------------------------------------------------------------------------------------
// QImaging TIFF
//---------------------------------------------------------------------------------------
int TMetaFormatManager::parse_qimagingtiff() {
  if (tagList == NULL) return false;
  if (tagList->count == 0) return false;
  if (tagList->tags == NULL) return false;  
  int pos;    

  /*
[Image Description]
Exposure: 000 : 00 : 00 . 300 : 000
Binning: 2 x 2
Gain: 2.000000
%Accumulated%=0

[Software]
QCapture Pro

[Date Time]
08/28/2006 04:34:47.000 PM
  */

  // check if it's QImage tiff file
  // should exist private tags 50288 and 50296
  if ((pos = tagPos(tagList, DIM_META_TIFF_TAG, 50288)) == -1) return false;
  if ((pos = tagPos(tagList, DIM_META_TIFF_TAG, 50296)) == -1) return false;

  // tag 305 should be "QCapture Pro"
  if ((pos = tagPos(tagList, DIM_META_TIFF_TAG, 305)) == -1) return false;
  int size = min(12, tagList->tags[pos].tagLength);
  if (strncmp((char *) tagList->tags[pos].tagData, "QCapture Pro", size) != 0) return false;

  // ok, we're sure it's QImaging, now read date/time

  // read tag 306 - Date/Time
  if ((pos = tagPos(tagList, DIM_META_TIFF_TAG, 306)) != -1) {
    int y=0, m=0, d=0, h=0, mi=0, s=0, ms=0;
    char ampm=0;
    //08/28/2006 04:34:47.000 PM
    if (tagList->tags[pos].tagData != NULL)
      SSCANF( (char *)tagList->tags[pos].tagData, "%d/%d/%d %d:%d:%d.%d %c", &m, &d, &y, &h, &mi, &s, &ms, &ampm );
    if (ampm == 'P') h += 12;
    SPRINTF(imaging_time, "%.4d-%.2d-%.2d %.2d:%.2d:%.2d", y, m, d, h, mi, s);
  }

  return true;
}


//---------------------------------------------------------------------------------------
// Generic TIFF
//---------------------------------------------------------------------------------------
void TMetaFormatManager::parse_gentiff() {
  if (tagList == NULL) return;
  if (tagList->count == 0) return;
  if (tagList->tags == NULL) return;  
  int pos;    

  if (parse_qimagingtiff()) return;

  // read tag 306 - Date/Time
  if ((pos = tagPos(tagList, DIM_META_TIFF_TAG, 306)) != -1) {

    int size = min(19, tagList->tags[pos].tagLength);
    strncpy(imaging_time, (char *) tagList->tags[pos].tagData, size);
    imaging_time[19] = 0;
  }
}

//---------------------------------------------------------------------------------------
// PSIA
//---------------------------------------------------------------------------------------
void TMetaFormatManager::parse_psia() {
  if (tagList == NULL) return;
  if (tagList->count == 0) return;
  if (tagList->tags == NULL) return;  
  //int pos;    

  // the x and y are parsed inside the format and passed with the image

  /*
  if (isTagPresent( tagList, DIM_META_TIFF_TAG, 50432 ) == TRUE) {
    
    if ((pos = tagPos(tagList, DIM_META_TIFF_TAG, 50435)) != -1)  {
      unsigned char *p = (unsigned char *) tagList->tags[pos].tagData;
      
      double v1 = * (double *) (p + DIM_PSIA_OFFSET_XSCANSIZE);
      if (dimBigendian) dimSwapDouble( &v1 );
      DIM_UINT32 v2 = * (DIM_UINT32 *) (p + DIM_PSIA_OFFSET_WIDTH);
      if (dimBigendian) dimSwapLong( &v2 );
      pixel_size[0] = v1 / (double) v2;

      //pixel_size[0] = * (double *) (p + DIM_PSIA_OFFSET_XSCANSIZE);
      //pixel_size[0] /= (double) * (DIM_UINT32 *) (p + DIM_PSIA_OFFSET_WIDTH);

      v1 = * (double *) (p + DIM_PSIA_OFFSET_YSCANSIZE);
      if (dimBigendian) dimSwapDouble( &v1 );
      v2 = * (DIM_UINT32 *) (p + DIM_PSIA_OFFSET_HEIGHT);
      if (dimBigendian) dimSwapLong( &v2 );
      pixel_size[1] = v1 / (double) v2;
      //pixel_size[1] = * (double *) (p + DIM_PSIA_OFFSET_YSCANSIZE);
      //pixel_size[1] /= (double) * (DIM_UINT32 *) (p + DIM_PSIA_OFFSET_HEIGHT);


      sprintf(imaging_time, "0000-00-00 00:00:00");
    }

    return;
  }
  */
}


//---------------------------------------------------------------------------------------
// Nanoscope
//---------------------------------------------------------------------------------------
void TMetaFormatManager::parse_nanoscope() {

  if (meta_data_text.size() <= 0) return;

  // get date and time
  // \Date: 06:32:07 PM Thu Dec 09 2004

  char *line = (char*) strstr( meta_data_text.c_str(), "Date: " );
  if (line != NULL) {
    int h=0, mi=0, s=0;
    int m=0, d=0, y=0;
    char t[1024];
    char month[3];

    sscanf( line, "Date: %d:%d:%d %2s %3s %3s %d %d", &h, &mi, &s, t, t, month, &d, &y ); 
    m = getMonthNum( month );
    SPRINTF(imaging_time, "%.4d-%.2d-%.2d %.2d:%.2d:%.2d", y, m, d, h, mi, s);
  }

}


//---------------------------------------------------------------------------------------
// Nanoscope
//---------------------------------------------------------------------------------------
void TMetaFormatManager::parse_ibw() {

  if (meta_data_text.size() <= 0) return;

//Date: Tue, Sep 20, 2005
//Time: 10:58:49 AM

  int h=0, mi=0, s=0;
  int m=0, d=0, y=0;

  char *line = (char*) strstr( meta_data_text.c_str(), "Date: " );
  if (line != NULL) {
    char t[1024];
    char month[3];

    sscanf( line, "Date: %3s, %3s %d, %d", t, month, &d, &y ); 
    m = getMonthNum( month );
  }

  line = (char*) strstr( meta_data_text.c_str(), "Time: " );
  if (line != NULL) {
    SSCANF( line, "Time: %d:%d:%d", &h, &mi, &s ); 
  }

  SPRINTF(imaging_time, "%.4d-%.2d-%.2d %.2d:%.2d:%.2d", y, m, d, h, mi, s);
}


//---------------------------------------------------------------------------------------
// PARSING
//---------------------------------------------------------------------------------------

void TMetaFormatManager::parse_meta_data() {

  if ( sessionIsCurrentFormat("STK") ) 
    parse_stk();

  if ( sessionIsCurrentFormat("PSIA") ) 
    parse_psia();

  if ( sessionIsCurrentFormat("FLUOVIEW") ) 
    parse_fluoview();

  if ( sessionIsCurrentFormat("LSM") ) 
    parse_lsm();

  if ( sessionIsCurrentFormat("TIFF") ) 
    parse_gentiff();

  if ( sessionIsCurrentFormat("BIORAD-PIC") ) 
    parse_biorad();

  if ( sessionIsCurrentFormat("NANOSCOPE") ) 
    parse_nanoscope();

  if ( sessionIsCurrentFormat("IBW") ) 
    parse_ibw();
}

//***********************************************************************

#undef min


