/*****************************************************************************
  Carl Zeiss LSM IO 
  Copyright (c) 2006 by Dmitry V. Fedorov <www.dimin.net> <dima@dimin.net>

  IMPLEMENTATION
  
  Programmer: Dima V. Fedorov <mailto:dima@dimin.net> <http://www.dimin.net/>
    
  History:
    03/29/2004 22:23 - First creation
        
  Ver : 1
*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "dim_tiff_format.h"
#include "dim_xtiffio.h"
#include "xstring.h"

void read_text_tag(TIFF *tif, TDimTiffIFD *ifd, DIM_UINT tag, MemIOBuf *outIOBuf);

//----------------------------------------------------------------------------
// PSIA MISC FUNCTIONS
//----------------------------------------------------------------------------

bool lsmIsTiffValid(TIFF *tif) {
  char *b_list = NULL;
  int16   d_list_count;
  int res[3] = {0,0,0};

  if (tif == 0) return FALSE;
  res[0] = TIFFGetField(tif, TIFFTAG_CZ_LSMINFO, &d_list_count, &b_list);

  if (res[0] == 1) return TRUE;
  return FALSE;
}


bool lsmIsTiffValid(TDimTiffParams *tiffParams) {
  if (tiffParams == NULL) return FALSE;
  if (isTagPresentInFirstIFD( &tiffParams->ifds, TIFFTAG_CZ_LSMINFO ) == TRUE) return TRUE;
  return FALSE;
}

void doSwabLSMINFO(CZ_LSMINFO *b) {
  
  TIFFSwabArrayOfLong( (uint32*) &b->u32MagicNumber, 10 );
  TIFFSwabArrayOfDouble( (double*) &b->f64VoxelSizeX, 3 );
  TIFFSwabArrayOfLong( (uint32*) &b->u32ScanType, 6 );
  TIFFSwabDouble( (double*) &b->f64TimeIntervall );
  TIFFSwabArrayOfLong( (uint32*) &b->u32OffsetChannelDataTypes, 8 );
}

void doSwabLSMCOLORS(CZ_ChannelColors *b) {
  TIFFSwabArrayOfLong( (uint32*) &b->s32BlockSize, 6 );
}

int lsmGetInfo (TDimTiffParams *tiffParams) {
  if (tiffParams == NULL) return 1;
  if (tiffParams->dimTiff == NULL) return 1;
  if (tiffParams->ifds.count <= 0) return 1;

  TDimImageInfo *info = &tiffParams->info;
  TDimLsmInfo *lsm = &tiffParams->lsmInfo;
  CZ_LSMINFO *lsmi = &lsm->lsm_info;


  lsm->pages_tiff = tiffParams->info.number_pages;
  lsm->pages = tiffParams->info.number_pages / 2;
  tiffParams->info.number_pages = lsm->pages;

  lsm->ch = tiffParams->info.samples;
  if (tiffParams->info.samples > 1) 
    tiffParams->info.imageMode = DIM_RGB;
  else
    tiffParams->info.imageMode = DIM_GRAYSCALE;    

  uint16 bitspersample = 1;  
  TIFFGetField(tiffParams->dimTiff, TIFFTAG_BITSPERSAMPLE, &bitspersample);  
  tiffParams->info.depth = bitspersample;

  // ------------------------------------------
  // get LSM INFO STRUCTURE
  // ------------------------------------------

  DIM_UCHAR *buf = NULL;
  uint32 size, type;
  if (!isTagPresentInFirstIFD( &tiffParams->ifds, TIFFTAG_CZ_LSMINFO )) return 1;
  readTiffTag (tiffParams->dimTiff, &tiffParams->ifds.ifds[0], TIFFTAG_CZ_LSMINFO, size, type, &buf);
  if ( (size <= 0) || (buf == NULL) ) return 1;
  if (bigendian) doSwabLSMINFO((CZ_LSMINFO *) buf);

  lsm->lsm_info = * (CZ_LSMINFO *) buf;
  freeTiffTagBuf( &buf );

  //
  memset( &lsm->lsm_colors, 0, sizeof(CZ_ChannelColors) );
  if (lsm->lsm_info.u32OffsetChannelColors > 0) {
    readTiffBufNoAlloc( tiffParams->dimTiff, lsm->lsm_info.u32OffsetChannelColors, sizeof(CZ_ChannelColors), TIFF_LONG, (unsigned char *) &lsm->lsm_colors );
    // swaped by readTiffBufNoAlloc
    //if (bigendian) doSwabLSMCOLORS( &lsm->lsm_colors );

    // here we have to assign channel names reading them from the file, but never seen this used...
  }

  //---------------------------------------------------------------
  // retreive meta-data
  //---------------------------------------------------------------
  lsm->ch       = lsmi->s32DimensionChannels;
  lsm->t_frames = lsmi->s32DimensionTime;
  lsm->z_slices = lsmi->s32DimensionZ;

  lsm->res[0] = lsmi->f64VoxelSizeX; // x in meters
  lsm->res[1] = lsmi->f64VoxelSizeY; // y in meters
  lsm->res[2] = lsmi->f64VoxelSizeZ; // z in meters
  lsm->res[3] = lsmi->f64TimeIntervall; // t in seconds

  //---------------------------------------------------------------
  // define dims
  //---------------------------------------------------------------
  info->number_z = lsm->z_slices;
  info->number_t = lsm->t_frames;

  if (lsm->z_slices > 1) {
    info->number_dims = 4;
    info->dimensions[3].dim = DIM_DIM_Z;
  }

  if (lsm->t_frames > 1) {
    info->number_dims = 4;
    info->dimensions[3].dim = DIM_DIM_T;
  }

  if ((lsm->z_slices > 1) && (lsm->t_frames > 1)) {
    info->number_dims = 5;
    info->dimensions[3].dim = DIM_DIM_Z;        
    info->dimensions[4].dim = DIM_DIM_T;
  }

  return 0;
}


//----------------------------------------------------------------------------
// READ/WRITE FUNCTIONS
//----------------------------------------------------------------------------



//----------------------------------------------------------------------------
// METADATA TEXT FUNCTIONS
//----------------------------------------------------------------------------

void lsmWriteTextMeta(TDimTiffParams *tiffParams, MemIOBuf *outIOBuf) {
  if (tiffParams == NULL) return;
  TDimLsmInfo *lsm = &tiffParams->lsmInfo;
  CZ_LSMINFO *lsmi = &lsm->lsm_info;

  xstring str;
  str += "[Carl Zeiss LSM Metadata]\n";

  str.saddf( "Number of channels:    %d\n", lsm->ch );
  str.saddf( "Number of time frames: %d\n", lsm->t_frames );
  str.saddf( "Number of Z slices:    %d\n", lsm->z_slices );
  str += "\n";    
  str.saddf( "X Resolution: %.6f microns\n", lsm->res[0]*1000000 );
  str.saddf( "Y Resolution: %.6f microns\n", lsm->res[1]*1000000 );
  str.saddf( "Z Resolution: %.6f microns\n", lsm->res[2]*1000000 );
  str.saddf( "T Resolution: %.6f seconds\n", lsm->res[3] );

  MemIO_WriteProc( (thandle_t) outIOBuf, (tdata_t) str.c_str(), str.size() );
}
























