/*****************************************************************************
  TIFF STK IO 
  Copyright (c) 2004 by Dmitry V. Fedorov <www.dimin.net> <dima@dimin.net>

  IMPLEMENTATION
  
  Programmer: Dima V. Fedorov <mailto:dima@dimin.net> <http://www.dimin.net/>
    
  TODO:
    At this point it's not supporting more then one sample per pixel...
    Check compression related to stripes, we might change it later

  History:
    03/29/2004 22:23 - First creation
    09/28/2005 23:10 - fixed bugs for bigendian architecture
        
  Ver : 2
*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <string>

// Disables Visual Studio 2005 warnings for deprecated code
#if ( defined(_MSC_VER) && (_MSC_VER >= 1400) )
  #pragma warning(disable:4996)
#endif 

#include "memio.h"
#include "dim_tiff_format.h"
#include "dim_stk_format.h"

#define HOURS_PER_DAY 24
#define MINS_PER_HOUR 60
#define SECS_PER_MIN  60
#define MSECS_PER_SEC 1000
#define MINS_PER_DAY  (HOURS_PER_DAY * MINS_PER_HOUR)
#define SECS_PER_DAY  (MINS_PER_DAY * SECS_PER_MIN)
#define MSECS_PER_DAY (SECS_PER_DAY * MSECS_PER_SEC)

//----------------------------------------------------------------------------
// STK MISC FUNCTIONS
//----------------------------------------------------------------------------
void JulianToYMD(unsigned long julian, unsigned short& year, unsigned char& month, unsigned char& day)
{
  long  a, b, c, d, e, alpha;
  long  z = julian + 1;

  // dealing with Gregorian calendar reform
  if (z < 2299161L) {
    a = z;
  } else {
    alpha = (long) ((z - 1867216.25) / 36524.25);
    a = z + 1 + alpha - alpha / 4;
  }

  b = ( a > 1721423L ? a + 1524 : a + 1158 );
  c = (long) ((b - 122.1) / 365.25);
  d = (long) (365.25 * c);
  e = (long) ((b - d) / 30.6001);

  day   = (unsigned char) (b - d - (long)(30.6001 * e));
  month = (unsigned char) ((e < 13.5) ? e - 1 : e - 13) - 1; // dima: sub 1 to match Date/Time TIFF tag
  year  = (short)         ((month > 2.5 ) ? (c - 4716) : c - 4715);
}

unsigned long YMDToJulian(unsigned short year, unsigned char month, unsigned char day)
{
  short a, b = 0;
  short work_year = year;
  short work_month = month;
  short work_day = day;

  // correct for negative year
  if (work_year < 0) {
    work_year++;
  }

  if (work_month <= 2) {
    work_year--;
    work_month += (short)12;
  }

  // deal with Gregorian calendar
  if (work_year*10000. + work_month*100. + work_day >= 15821015.) {
    a = (short)(work_year/100.);
    b = (short)(2 - a + a/4);
  }

  return  (long) (365.25*work_year) +
          (long) (30.6001 * (work_month+1)) +
          work_day + 1720994L + b;
}

void divMod(int Dividend, int Divisor, int &Result, int &Remainder)
{
  Result    = (int) ( ((double) Dividend) / ((double) Divisor) );
  Remainder = Dividend % Divisor;
}

// this convert miliseconds since midnight into hour/minute/second
void MiliMidnightToHMS(unsigned long ms, unsigned char& hour, unsigned char& minute, unsigned char& second)
{
  int h, m, s, fms;
  int mc, msc;

  divMod(ms,  SECS_PER_MIN * MSECS_PER_SEC, mc, msc);
  divMod(mc,  MINS_PER_HOUR, h, m);
  divMod(msc, MSECS_PER_SEC, s, fms);

  hour = h;
  minute = m;
  second = s;
}


//----------------------------------------------------------------------------
// STK UTIL FUNCTIONS
//----------------------------------------------------------------------------

bool stkAreValidParams(TDimTiffParams *tiffParams)
{
  if (tiffParams == NULL) return FALSE;
  if (tiffParams->dimTiff == NULL) return FALSE;

  return TRUE;
}

long stkGetNumPlanes(TIFF *tif)
{
  double *d_list = NULL;
  long *l_list = NULL;
  int16   d_list_count[3];
  int res[3] = {0,0,0};

  if (tif == 0) return 0;

  res[0] = TIFFGetField(tif, 33629, &d_list_count[0], &d_list);
  res[1] = TIFFGetField(tif, 33630, &d_list_count[1], &d_list);
  res[2] = TIFFGetField(tif, 33631, &d_list_count[2], &l_list);

  // if tag 33629 exists then the file is valid STAK file
  if (res[0] == 1) return d_list_count[0];
  if (res[1] == 1) return d_list_count[1];
  if (res[2] == 1) return d_list_count[2];

  return 1;
}

bool stkIsTiffValid(TIFF *tif)
{
  double *d_list = NULL;
  long *l_list = NULL;
  int16   d_list_count;
  int res[3] = {0,0,0};

  if (tif == 0) return FALSE;

  res[0] = TIFFGetField(tif, 33629, &d_list_count, &d_list);
  res[1] = TIFFGetField(tif, 33630, &d_list_count, &d_list);
  res[2] = TIFFGetField(tif, 33631, &d_list_count, &l_list);

  // if tag 33629 exists then the file is valid STAK file
  if (res[0] == 1) return TRUE;
  if (res[1] == 1) return TRUE;
  if (res[2] == 1) return TRUE;

  return FALSE;
}

bool stkIsTiffValid(TDimTiffParams *tiffParams)
{
  if (tiffParams == NULL) return FALSE;

  // if tag 33629 exists then the file is valid STAK file
  if (isTagPresentInFirstIFD( &tiffParams->ifds, 33629 ) == TRUE) return TRUE;
  if (isTagPresentInFirstIFD( &tiffParams->ifds, 33630 ) == TRUE) return TRUE;
  if (isTagPresentInFirstIFD( &tiffParams->ifds, 33631 ) == TRUE) return TRUE;

  return FALSE;
}

void stkGetOffsets(TIFF *tif, TDimTiffParams *tiffParams )
{
  /*
  the default libtiff strip reading is not good enough since it only gives us strip offsets for 
  the first plane although all the offsets for all the planes are actually stored
  so for uncompressed images use default and for LZW the custom methods
  */
  
  TDimStkInfo *stkInfo = &tiffParams->stkInfo;
  stkInfo->strips_per_plane = TIFFNumberOfStrips( tif );

  if (stkInfo->strip_offsets != NULL) _TIFFfree(stkInfo->strip_offsets);
  stkInfo->strip_offsets = NULL;
  if (stkInfo->strip_bytecounts != NULL) _TIFFfree(stkInfo->strip_bytecounts);
  stkInfo->strip_bytecounts = NULL;

  if (tiffParams == NULL) return;
  if (tiffParams->dimTiff == NULL) return;
  if (tiffParams->subType != tstStk) return;

  TDimTiffIFD *ifd = &tiffParams->ifds.ifds[0];
  int N = tiffParams->stkInfo.metaData.N;
  uint32 tag = 0;

  // -----------------------------
  // read Strip Offsets
  // -----------------------------
  tag = 273;
  if (isTagPresentInFirstIFD( &tiffParams->ifds, tag ) == TRUE)
  {
    uint32 *buf = NULL;
    uint32 size=0, type=0;

    readTiffTag ( tif, ifd, tag, size, type, (DIM_UCHAR **) &buf );
    if (buf != NULL) 
    {
      stkInfo->strip_offsets = (uint32 *) _TIFFmalloc( size );
      _TIFFmemcpy( stkInfo->strip_offsets, buf, size );
      _TIFFfree( buf );
    }
  } // strip offsets
  
  // -----------------------------
  // read strip_bytecounts also
  // -----------------------------
  tag = 279;
  if (isTagPresentInFirstIFD( &tiffParams->ifds, tag ) == TRUE)
  {
    uint32 *buf = NULL;
    uint32 size=0, type=0;

    readTiffTag ( tif, ifd, tag, size, type, (DIM_UCHAR **) &buf );
    if (buf != NULL) 
    {
      stkInfo->strip_bytecounts = (uint32 *) _TIFFmalloc( size );
      _TIFFmemcpy( stkInfo->strip_bytecounts, buf, size );
      _TIFFfree( buf );
    }
  } // strip_bytecounts
}

void stkClearOffsets(TDimStkInfo *stkInfo)
{
  if ( (stkInfo->strip_offsets != NULL) && (stkInfo->strips_per_plane > 0) )
    _TIFFfree(stkInfo->strip_offsets);

  if ( (stkInfo->strip_bytecounts != NULL) && (stkInfo->strips_per_plane > 0) )
    _TIFFfree(stkInfo->strip_bytecounts);

  stkInfo->strip_offsets    = NULL;
  stkInfo->strip_bytecounts = NULL;
  stkInfo->strips_per_plane = 0;
}


//----------------------------------------------------------------------------
// PARSE UIC TAGS
//----------------------------------------------------------------------------

// UIC2Tag - 33629  
//At the indicated offset, there is a table of 6*N LONG values 
//indicating for each plane in order: 
//Z distance (numerator), Z distance (denominator), 
//creation date, creation time, modification date, modification time.
void stkParseUIC2Tag( TDimTiffParams *tiffParams )
{
  if (tiffParams == NULL) return;
  if (tiffParams->dimTiff == NULL) return;
  if (tiffParams->subType != tstStk) return;

  TDimStkInfo *stkInfo = &tiffParams->stkInfo;
  TDimTiffIFD *ifd = &tiffParams->ifds.ifds[0];

  // read UIC2
  int tag = 33629;
  if (isTagPresentInFirstIFD( &tiffParams->ifds, tag ) != TRUE) return;
  else
  {
    long i;
    long N = stkInfo->metaData.N;
    DIM_UINT32 *buf = NULL;
    readTiffCustomTag (tiffParams->dimTiff, ifd, tag, 6*N*sizeof(DIM_INT32), DIM_TAG_LONG, (DIM_UCHAR **) &buf);
    if (buf == NULL) return;

    for (i=0; i<N; i++)
    {
      stkInfo->metaData.zDistance[i].num    = buf[i*6+0];
      stkInfo->metaData.zDistance[i].den    = buf[i*6+1];
      stkInfo->metaData.creationDate[i]     = buf[i*6+2];
      stkInfo->metaData.creationTime[i]     = buf[i*6+3];
      stkInfo->metaData.modificationDate[i] = buf[i*6+4];
      stkInfo->metaData.modificationTime[i] = buf[i*6+5];
    }

    freeTiffTagBuf( (DIM_UCHAR **) &buf );
  }
}


// UIC3Tag - 33630 
//A table of 2*N LONG values indicating for each plane in order: 
//wavelength (numerator), wavelength (denominator).
void stkParseUIC3Tag( TDimTiffParams *tiffParams )
{
  if (tiffParams == NULL) return;
  if (tiffParams->dimTiff == NULL) return;
  if (tiffParams->subType != tstStk) return;

  TDimStkInfo *stkInfo = &tiffParams->stkInfo;
  TDimTiffIFD *ifd = &tiffParams->ifds.ifds[0];

  // read UIC3
  int tag = 33630;
  if (isTagPresentInFirstIFD( &tiffParams->ifds, tag ) != TRUE) return;
  else
  {
    long i=0;
    long N = tiffParams->stkInfo.metaData.N;
    DIM_UINT32 *buf = NULL;
 
    readTiffCustomTag (tiffParams->dimTiff, ifd, tag, 2*N*sizeof(DIM_UINT32), DIM_TAG_LONG, (DIM_UCHAR **) &buf);
    if (buf == NULL) return;

    for (i=0; i<N; i++)
    {
      stkInfo->metaData.wavelength[i].num = buf[i*2];
      stkInfo->metaData.wavelength[i].den = buf[i*2+1];
    }

    freeTiffTagBuf( (DIM_UCHAR **) &buf );
  }
}

//UIC1Tag - 33628 
//If this tag exists and is of type LONG, then at the indicated offset 
//there is a series of N pairs consisting of an ID code of size LONG and a 
//variable-sized (ID-dependent) block of data. The possible tags and their 
//values are defined in the "Tag ID Codes" section below. These values are 
//used internally by the Meta Imaging Series applications, and may not be 
//useful to other applications. To replicate the behavior of MetaMorph, 
//this table should be read and its values stored after the table 
//indicated by UIC4Tag is read.
// NOTE: difference form tag UIC4 is that this one contains exactly N pairs
// and long organization of data... We're not reading tag UIC4 at all at the moment

void stkReadStringEntry(long offset, TDimTiffParams *tiffParams, char **string)
{
  //TDimStkInfo *stkInfo = &tiffParams->stkInfo;
  TIFF *tif = tiffParams->dimTiff;
  if (offset <= 0) return;

  long size;
  readTiffBufNoAlloc (tif, offset, sizeof( DIM_LONG ), DIM_TAG_LONG, (DIM_UCHAR *) &size);
  *string = (char *) _TIFFmalloc( size+1 );
  (*string)[size] = '\0';
  readTiffBufNoAlloc (tif, offset+sizeof( DIM_LONG ), size, DIM_TAG_LONG, (DIM_UCHAR *) (*string));
}

void stkParseIDEntry(long *pair, long offset, TDimTiffParams *tiffParams)
{
  TDimStkInfo *stkInfo = &tiffParams->stkInfo;
  //TDimTiffIFD *ifd = &tiffParams->ifds.ifds[0];
  TIFF *tif = tiffParams->dimTiff;

  if (pair[0] == DIM_STK_AutoScale)
    stkInfo->metaData.AutoScale = pair[1];


  if (pair[0] == DIM_STK_MinScale)
    stkInfo->metaData.MinScale = pair[1];


  if (pair[0] == DIM_STK_MaxScale)
    stkInfo->metaData.MaxScale = pair[1];


  if (pair[0] == DIM_STK_SpatialCalibration)
    stkInfo->metaData.SpatialCalibration = pair[1];


  if (pair[0] == DIM_STK_XCalibration)
    readTiffBufNoAlloc (tif, pair[1], 2*sizeof( DIM_LONG ), DIM_TAG_LONG, (DIM_UCHAR *) stkInfo->metaData.XCalibration);

  if (pair[0] == DIM_STK_YCalibration)
    readTiffBufNoAlloc (tif, pair[1], 2*sizeof( DIM_LONG ), DIM_TAG_LONG, (DIM_UCHAR *) stkInfo->metaData.YCalibration);

  if (pair[0] == DIM_STK_CalibrationUnits)
    stkReadStringEntry(pair[1], tiffParams, (char **) &stkInfo->metaData.CalibrationUnits);

  if (pair[0] == DIM_STK_Name)
    stkReadStringEntry(pair[1], tiffParams, (char **) &stkInfo->metaData.Name);

  if (pair[0] == DIM_STK_ThreshState)
    stkInfo->metaData.ThreshState = pair[1];

  if (pair[0] == DIM_STK_ThreshStateRed)
    stkInfo->metaData.ThreshStateRed = pair[1];

  if (pair[0] == DIM_STK_ThreshStateGreen)
    stkInfo->metaData.ThreshStateGreen = pair[1];

  if (pair[0] == DIM_STK_ThreshStateBlue)
    stkInfo->metaData.ThreshStateBlue = pair[1];

  if (pair[0] == DIM_STK_ThreshStateLo)
    stkInfo->metaData.ThreshStateLo = pair[1];

  if (pair[0] == DIM_STK_ThreshStateHi)
    stkInfo->metaData.ThreshStateHi = pair[1];

  if (pair[0] == DIM_STK_Zoom)
    stkInfo->metaData.Zoom = pair[1];

  if (pair[0] == DIM_STK_CreateTime)
    readTiffBufNoAlloc (tif, pair[1], 2*sizeof( DIM_LONG ), DIM_TAG_LONG, (DIM_UCHAR *) stkInfo->metaData.CreateTime);

  if (pair[0] == DIM_STK_LastSavedTime)
    readTiffBufNoAlloc (tif, pair[1], 2*sizeof( DIM_LONG ), DIM_TAG_LONG, (DIM_UCHAR *) stkInfo->metaData.LastSavedTime);

  if (pair[0] == DIM_STK_currentBuffer)
    stkInfo->metaData.currentBuffer = pair[1];

  if (pair[0] == DIM_STK_grayFit)
    stkInfo->metaData.grayFit = pair[1];

  if (pair[0] == DIM_STK_grayPointCount)
    stkInfo->metaData.grayPointCount = pair[1];

  if (pair[0] == DIM_STK_grayX)
    readTiffBufNoAlloc (tif, pair[1], 2*sizeof( DIM_LONG ), DIM_TAG_LONG, (DIM_UCHAR *) stkInfo->metaData.grayX);

  if (pair[0] == DIM_STK_grayY)
    readTiffBufNoAlloc (tif, pair[1], 2*sizeof( DIM_LONG ), DIM_TAG_LONG, (DIM_UCHAR *) stkInfo->metaData.grayY);

  if (pair[0] == DIM_STK_grayMin)
    readTiffBufNoAlloc (tif, pair[1], 2*sizeof( DIM_LONG ), DIM_TAG_LONG, (DIM_UCHAR *) stkInfo->metaData.grayMin);

  if (pair[0] == DIM_STK_grayMax)
    readTiffBufNoAlloc (tif, pair[1], 2*sizeof( DIM_LONG ), DIM_TAG_LONG, (DIM_UCHAR *) stkInfo->metaData.grayMax);

  if (pair[0] == DIM_STK_grayUnitName)
    stkReadStringEntry(pair[1], tiffParams, (char **) &stkInfo->metaData.grayUnitName);

  if (pair[0] == DIM_STK_StandardLUT)
    stkInfo->metaData.StandardLUT = pair[1];

  if (pair[0] == DIM_STK_AutoScaleLoInfo)
    readTiffBufNoAlloc (tif, pair[1], 2*sizeof( DIM_LONG ), DIM_TAG_LONG, (DIM_UCHAR *) stkInfo->metaData.AutoScaleLoInfo);

  if (pair[0] == DIM_STK_AutoScaleHiInfo)
    readTiffBufNoAlloc (tif, pair[1], 2*sizeof( DIM_LONG ), DIM_TAG_LONG, (DIM_UCHAR *) stkInfo->metaData.AutoScaleHiInfo);

  if (pair[0] == DIM_STK_Gamma)
    readTiffBufNoAlloc (tif, pair[1], sizeof( DIM_LONG ), DIM_TAG_LONG, (DIM_UCHAR *) stkInfo->metaData.Gamma);

  if (pair[0] == DIM_STK_GammaRed)
    readTiffBufNoAlloc (tif, pair[1], sizeof( DIM_LONG ), DIM_TAG_LONG, (DIM_UCHAR *) stkInfo->metaData.GammaRed);

  if (pair[0] == DIM_STK_GammaGreen)
    readTiffBufNoAlloc (tif, pair[1], sizeof( DIM_LONG ), DIM_TAG_LONG, (DIM_UCHAR *) stkInfo->metaData.GammaGreen);

  if (pair[0] == DIM_STK_GammaBlue)
    readTiffBufNoAlloc (tif, pair[1], sizeof( DIM_LONG ), DIM_TAG_LONG, (DIM_UCHAR *) stkInfo->metaData.GammaBlue);
}

void stkParseUIC1Tag( TDimTiffParams *tiffParams )
{
  if (tiffParams == NULL) return;
  if (tiffParams->dimTiff == NULL) return;
  if (tiffParams->subType != tstStk) return;

  TDimTiffIFD *ifd = &tiffParams->ifds.ifds[0];
  TIFF *tif = tiffParams->dimTiff;

  int tag = 33628;
  if (isTagPresentInFirstIFD( &tiffParams->ifds, tag ) != TRUE) return;
  
  long i, offset;
  long N = tiffParams->stkInfo.metaData.N; 
  long pair[2];

  offset = getTiffTagOffset(tif, ifd, tag);
  if (offset == -1) return;
  long num_ids = getTiffTagCount(tif, ifd, tag);
  long id_offset = offset;

  // now read and parce ID table
  for (i=0; i<num_ids; i++)
  {
    readTiffBufNoAlloc( tif, id_offset, 2*sizeof( DIM_LONG ), DIM_TAG_LONG, (DIM_UCHAR *) pair);
    stkParseIDEntry(pair, id_offset, tiffParams);
    id_offset += 2*sizeof( DIM_LONG );
  }

}

void stkParseUICTags( TDimTiffParams *tiffParams )
{
  if (tiffParams == NULL) return;
  if (tiffParams->dimTiff == NULL) return;
  if (tiffParams->subType != tstStk) return;
  
  stkParseUIC2Tag( tiffParams );
  stkParseUIC3Tag( tiffParams );
  stkParseUIC1Tag( tiffParams );
}


//----------------------------------------------------------------------------
// STK INFO
//----------------------------------------------------------------------------

void initRational(TDimStkRational *r, long N)
{
  int i;
  if (r == NULL) return;

  for (i=0; i<N; i++) {
    r[i].num = 0;
    r[i].den = 1;
  }
}

void initLong(long *r, long N)
{
  int i;
  if (r == NULL) return;

  for (i=0; i<N; i++) r[i] = 0;
}

void stkAllocMetaInfo( TDimStkInfo *stkInfo, DIM_LONG N )
{
  if (stkInfo == NULL) return;
  if (N <= 0) return;

  stkInfo->metaData.N = N;

  stkInfo->metaData.StagePositionX = (TDimStkRational *) _TIFFmalloc( N*sizeof(TDimStkRational) );
  initRational(stkInfo->metaData.StagePositionX, N );

  stkInfo->metaData.StagePositionY = (TDimStkRational *) _TIFFmalloc( N*sizeof(TDimStkRational) );
  initRational(stkInfo->metaData.StagePositionY, N );

  stkInfo->metaData.CameraChipOffsetX = (TDimStkRational *) _TIFFmalloc( N*sizeof(TDimStkRational) );
  initRational(stkInfo->metaData.CameraChipOffsetX, N );

  stkInfo->metaData.CameraChipOffsetY = (TDimStkRational *) _TIFFmalloc( N*sizeof(TDimStkRational) );
  initRational(stkInfo->metaData.CameraChipOffsetY, N );

  stkInfo->metaData.AbsoluteZ = (TDimStkRational *) _TIFFmalloc( N*sizeof(TDimStkRational) );
  initRational(stkInfo->metaData.AbsoluteZ, N );

  stkInfo->metaData.AbsoluteZValid = (TDimStkRational *) _TIFFmalloc( N*sizeof(TDimStkRational) );
  initRational(stkInfo->metaData.AbsoluteZValid, N );

  stkInfo->metaData.wavelength = (TDimStkRational *) _TIFFmalloc( N*sizeof(TDimStkRational) );
  initRational(stkInfo->metaData.wavelength, N );

  stkInfo->metaData.zDistance = (TDimStkRational *) _TIFFmalloc( N*sizeof(TDimStkRational) );
  initRational(stkInfo->metaData.zDistance, N );

  stkInfo->metaData.creationDate = (DIM_LONG *) _TIFFmalloc( N*sizeof(DIM_LONG) );
  initLong(stkInfo->metaData.creationDate, N );

  stkInfo->metaData.creationTime = (DIM_LONG *) _TIFFmalloc( N*sizeof(DIM_LONG) );
  initLong(stkInfo->metaData.creationTime, N );

  stkInfo->metaData.modificationDate = (DIM_LONG *) _TIFFmalloc( N*sizeof(DIM_LONG) );
  initLong(stkInfo->metaData.modificationDate, N );

  stkInfo->metaData.modificationTime = (DIM_LONG *) _TIFFmalloc( N*sizeof(DIM_LONG) );
  initLong(stkInfo->metaData.modificationTime, N );

  #if defined(DEBUG) || defined(_DEBUG)
  printf("STK: Allocated data for %d pages\n", N);  
  #endif  
}

int stkGetInfo ( TDimTiffParams *tiffParams )
{
  if (tiffParams == NULL) return 1;
  if (tiffParams->dimTiff == NULL) return 1;
  if (tiffParams->ifds.count <= 0) return 1;

  stkAllocMetaInfo( &tiffParams->stkInfo, stkGetNumPlanes( tiffParams->dimTiff ) );

  stkGetOffsets( tiffParams->dimTiff, tiffParams );

  stkParseUICTags( tiffParams );

  return 0;
}

void stkInitInfo( TDimStkInfo *stkInfo )
{
  if (stkInfo == NULL) return;

  stkInfo->strips_per_plane = 0;
  stkInfo->strip_offsets = NULL;
  stkInfo->strip_bytecounts = NULL;

  stkInfo->metaData.CalibrationUnits  = NULL;
  stkInfo->metaData.Name              = NULL;
  stkInfo->metaData.grayUnitName      = NULL;
  stkInfo->metaData.StagePositionX    = NULL;
  stkInfo->metaData.StagePositionY    = NULL;
  stkInfo->metaData.CameraChipOffsetX = NULL;
  stkInfo->metaData.CameraChipOffsetY = NULL;
  stkInfo->metaData.StageLabel        = NULL;
  stkInfo->metaData.AbsoluteZ         = NULL;
  stkInfo->metaData.AbsoluteZValid    = NULL;
  stkInfo->metaData.wavelength        = NULL;
  stkInfo->metaData.zDistance         = NULL;
  stkInfo->metaData.creationDate      = NULL;
  stkInfo->metaData.creationTime      = NULL;
  stkInfo->metaData.modificationDate  = NULL;
  stkInfo->metaData.modificationTime  = NULL;

  stkInfo->metaData.N                      = 0;
  stkInfo->metaData.AutoScale              = 0;
  stkInfo->metaData.MinScale               = 0;
  stkInfo->metaData.MaxScale               = 0;
  stkInfo->metaData.SpatialCalibration     = 0;
  stkInfo->metaData.XCalibration[0]        = 0;
  stkInfo->metaData.XCalibration[1]        = 1;
  stkInfo->metaData.YCalibration[0]        = 0;
  stkInfo->metaData.YCalibration[1]        = 1;
  stkInfo->metaData.ThreshState            = 0;
  stkInfo->metaData.ThreshStateRed         = 0;
  stkInfo->metaData.ThreshStateGreen       = 0;
  stkInfo->metaData.ThreshStateBlue        = 0;
  stkInfo->metaData.ThreshStateLo          = 0;
  stkInfo->metaData.ThreshStateHi          = 0;
  stkInfo->metaData.Zoom                   = 0;
  stkInfo->metaData.CreateTime[0]          = 0;
  stkInfo->metaData.CreateTime[1]          = 1;
  stkInfo->metaData.LastSavedTime[0]       = 0;
  stkInfo->metaData.LastSavedTime[1]       = 1;
  stkInfo->metaData.currentBuffer          = 0;
  stkInfo->metaData.grayFit                = 0;
  stkInfo->metaData.grayPointCount         = 0;
  stkInfo->metaData.grayX[0]               = 0;
  stkInfo->metaData.grayX[1]               = 1;
  stkInfo->metaData.grayY[0]               = 0;
  stkInfo->metaData.grayY[1]               = 1;
  stkInfo->metaData.grayMin[0]             = 0;
  stkInfo->metaData.grayMin[1]             = 1;
  stkInfo->metaData.grayMax[0]             = 0;
  stkInfo->metaData.grayMax[1]             = 1;
  stkInfo->metaData.StandardLUT            = 0;
  stkInfo->metaData.OverlayMask            = 0;
  stkInfo->metaData.OverlayCompress        = 0;
  stkInfo->metaData.Overlay                = 0;
  stkInfo->metaData.SpecialOverlayMask     = 0;
  stkInfo->metaData.SpecialOverlayCompress = 0;
  stkInfo->metaData.SpecialOverlay         = 0;
  stkInfo->metaData.ImageProperty          = 0;
  stkInfo->metaData.AutoScaleLoInfo[0]     = 0;
  stkInfo->metaData.AutoScaleLoInfo[1]     = 1;
  stkInfo->metaData.AutoScaleHiInfo[0]     = 0;
  stkInfo->metaData.AutoScaleHiInfo[1]     = 1;
  stkInfo->metaData.Gamma                  = 0;
  stkInfo->metaData.GammaRed               = 0;
  stkInfo->metaData.GammaGreen             = 0;
  stkInfo->metaData.GammaBlue              = 0;
}

void stkClearInfo ( TDimTiffParams *tiffParams )
{
  if (tiffParams == NULL) return;
  if (tiffParams->subType != tstStk) return;

  TDimStkInfo *stkInfo = &tiffParams->stkInfo;

  stkClearOffsets( stkInfo );
 
  if (  stkInfo->metaData.CalibrationUnits != NULL )
    _TIFFfree(stkInfo->metaData.CalibrationUnits);

  if (  stkInfo->metaData.Name != NULL )
    _TIFFfree(stkInfo->metaData.Name);

  if (  stkInfo->metaData.grayUnitName != NULL )
    _TIFFfree(stkInfo->metaData.grayUnitName);

  if (  stkInfo->metaData.StagePositionX != NULL )
    _TIFFfree(stkInfo->metaData.StagePositionX);

  if (  stkInfo->metaData.StagePositionY != NULL )
    _TIFFfree(stkInfo->metaData.StagePositionY);

  if (  stkInfo->metaData.CameraChipOffsetX != NULL )
    _TIFFfree(stkInfo->metaData.CameraChipOffsetX);

  if (  stkInfo->metaData.CameraChipOffsetY != NULL )
    _TIFFfree(stkInfo->metaData.CameraChipOffsetY); 

  if (  stkInfo->metaData.StageLabel != NULL )  
    _TIFFfree(stkInfo->metaData.StageLabel);

  if (  stkInfo->metaData.AbsoluteZ != NULL )
    _TIFFfree(stkInfo->metaData.AbsoluteZ);

  if (  stkInfo->metaData.AbsoluteZValid != NULL )
    _TIFFfree(stkInfo->metaData.AbsoluteZValid);

  if (  stkInfo->metaData.wavelength != NULL )
    _TIFFfree(stkInfo->metaData.wavelength);

  if (  stkInfo->metaData.zDistance != NULL )
    _TIFFfree(stkInfo->metaData.zDistance);

  if (  stkInfo->metaData.creationDate != NULL )
    _TIFFfree(stkInfo->metaData.creationDate);

  if (  stkInfo->metaData.creationTime != NULL )
    _TIFFfree(stkInfo->metaData.creationTime);

  if (  stkInfo->metaData.modificationDate != NULL )
    _TIFFfree(stkInfo->metaData.modificationDate);

  if (  stkInfo->metaData.modificationTime != NULL )
    _TIFFfree(stkInfo->metaData.modificationTime);

  stkInitInfo( stkInfo );
  return;
}


//----------------------------------------------------------------------------
// READ/WRITE FUNCTIONS
//----------------------------------------------------------------------------


DIM_UINT  stkReadPlane(TDimTiffParams *tiffParams, int plane, TDimImageBitmap *img, TDimFormatHandle *fmtHndl)
{
  if (tiffParams == 0) return 1;
  if (img        == 0) return 1;
  if (tiffParams->dimTiff == 0) return 1;
  DIM_UINT sample = 0;

  TIFF *tif = tiffParams->dimTiff;
  uint16 photometric = PHOTOMETRIC_MINISWHITE;
  uint16 compress_tag;
  uint16 bitspersample = 8;
  uint16 samplesperpixel = 1;
  uint32 height = 0; 
  uint32 width = 0;
  toff_t old_pos = tif->tif_curoff;
  uint32 rowsperstrip   = 1;

  (void) TIFFGetFieldDefaulted(tif, TIFFTAG_COMPRESSION, &compress_tag); 
  TIFFGetField(tif, TIFFTAG_SAMPLESPERPIXEL, &samplesperpixel);
  TIFFGetField(tif, TIFFTAG_BITSPERSAMPLE, &bitspersample);
  TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &width);
  TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &height);  
  TIFFGetField(tif, TIFFTAG_ROWSPERSTRIP, &rowsperstrip); 


  DIM_UINT Bpp = (unsigned int) ceil( ((double) img->i.depth) / 8.0 );
  long size = img->i.width * img->i.height * Bpp;

  long file_plane_offset = tiffParams->stkInfo.strip_offsets[0] + (size * plane);

  if (compress_tag != COMPRESSION_NONE) 
    file_plane_offset = tiffParams->stkInfo.strip_offsets[plane * tiffParams->stkInfo.strips_per_plane ]; 



  // position file position into begining of the plane
  if (tif->tif_seekproc((thandle_t) tif->tif_fd, file_plane_offset, SEEK_SET) != 0)
  {
    
    //---------------------------------------------------
    // load the buffer with decompressed plane data
    //---------------------------------------------------
  
    if (compress_tag == COMPRESSION_NONE) {
      dimProgress( fmtHndl, 0, 10, "Reading STK" );
      //if ( dimTestAbort( fmtHndl ) == 1) break; 
      
      int read_size = tif->tif_readproc((thandle_t) tif->tif_fd, img->bits[sample], size);
      if (read_size != size) return 1;
      // now swap bytes if needed
      tif->tif_postdecode(tif, (tidata_t) img->bits[sample], size); 
    }
    else
    {
      int strip_size = width * Bpp * samplesperpixel * rowsperstrip;
      DIM_UCHAR *buf = (DIM_UCHAR *) _TIFFmalloc( strip_size ); 

      // ---------------------------------------------------
      // let's tweak libtiff and change offsets and bytecounts for correct values for this plane
	    TIFFDirectory *td = &tif->tif_dir;

      uint32 *plane_strip_offsets = tiffParams->stkInfo.strip_offsets + (plane * tiffParams->stkInfo.strips_per_plane);
      //memcpy( td->td_stripoffset, plane_strip_offsets, tiffParams->stkInfo.strips_per_plane * sizeof(uint32) );
      _TIFFmemcpy( td->td_stripoffset, plane_strip_offsets, tiffParams->stkInfo.strips_per_plane * sizeof(uint32) );

      uint32 *plane_strip_bytecounts = tiffParams->stkInfo.strip_bytecounts + (plane * tiffParams->stkInfo.strips_per_plane);
      //memcpy( td->td_stripbytecount, plane_strip_bytecounts, tiffParams->stkInfo.strips_per_plane * sizeof(uint32) );
      _TIFFmemcpy( td->td_stripbytecount, plane_strip_bytecounts, tiffParams->stkInfo.strips_per_plane * sizeof(uint32) );
      // ---------------------------------------------------

      int strip_num = plane * tiffParams->stkInfo.strips_per_plane;
      unsigned int strip_offset = 0;
      while (strip_offset < tiffParams->stkInfo.strips_per_plane) {
        dimProgress( fmtHndl, strip_offset, tiffParams->stkInfo.strips_per_plane, "Reading STK" );
        if ( dimTestAbort( fmtHndl ) == 1) break;  

        int read_size = TIFFReadEncodedStrip (tif, strip_offset, (tdata_t) buf, (tsize_t) strip_size);
        if (read_size == -1) return 1;
        //memcpy( ((unsigned char *) img->bits[sample])+strip_offset*strip_size, buf, read_size );
        _TIFFmemcpy( ((unsigned char *) img->bits[sample])+strip_offset*strip_size, buf, read_size );
        ++strip_offset;
      }

      _TIFFfree( buf );  
    }

    // invert it if we got negative
    TIFFGetField(tiffParams->dimTiff, TIFFTAG_PHOTOMETRIC, &photometric);
    if (photometric == PHOTOMETRIC_MINISWHITE)
      invertSample(img, 0);
  }

  tif->tif_seekproc((thandle_t) tif->tif_fd, old_pos, SEEK_SET);

  return 0;
}


//----------------------------------------------------------------------------
// METADATA TEXT FUNCTIONS
//----------------------------------------------------------------------------

//----------------------------------------------------------------------------
// META DATA PROC
//----------------------------------------------------------------------------

DIM_UINT stkAddOneTag (TDimFormatHandle *fmtHndl, int tag, const char* str)
{
  DIM_UCHAR *buf = NULL;
  DIM_UINT32 buf_size = strlen(str);
  DIM_UINT32 buf_type = DIM_TAG_ASCII;

  if ( (buf_size == 0) || (str == NULL) ) return 1;
  else
  {
    // now add tag into structure
    TDimTagItem item;

    buf = (unsigned char *) DimMalloc(buf_size + 1);
    strncpy((char *) buf, str, buf_size);
    buf[buf_size] = '\0';

    item.tagGroup  = DIM_META_STK;
    item.tagId     = tag;
    item.tagType   = buf_type;
    item.tagLength = buf_size;
    item.tagData   = buf;

    addMetaTag( &fmtHndl->metaData, item);
  }

  return 0;
}


DIM_UINT stkReadMetaMeta (TDimFormatHandle *fmtHndl, int group, int tag, int type)
{
  if (fmtHndl == NULL) return 1;
  if (fmtHndl->internalParams == NULL) return 1;
  TDimTiffParams *tiffParams = (TDimTiffParams *) fmtHndl->internalParams;
  TDimStkMetaData *meta = &tiffParams->stkInfo.metaData;
  
  if ( (group != DIM_META_STK) && (group != -1) ) return 1;


  // add tag UIC2Tag 33629, some info
  std::string str_uic2 = "";
  char text[1024];

  int i;
  for (i=0; i<meta->N; i++)
  {
    unsigned short y;
    unsigned char m, d;
    unsigned char h, mi, s;

    JulianToYMD(meta->creationDate[i], y, m, d);
    MiliMidnightToHMS(meta->creationTime[i], h, mi, s);

    sprintf(text, "Page %.3d: %.4d-%.2d-%.2d %.2d:%.2d:%.2d\n", i, y, m, d, h, mi, s );
    str_uic2 += text;
  }
 
  if (str_uic2.size() > 0) 
    stkAddOneTag ( fmtHndl, 33629, str_uic2.c_str() );

  return 0;
}

void stkWriteTextMeta(TDimTiffParams *tiffParams, MemIOBuf *outIOBuf)
{
  if (tiffParams == NULL) return;
  if (tiffParams->subType != tstStk) return;
  if (outIOBuf == NULL) return;
  TDimStkMetaData *meta = &tiffParams->stkInfo.metaData;
  char text[1024];

/*
  //long CreateTime[2];
  //long LastSavedTime[2];
*/
  sprintf(text, "[MetaMorph Stack (STK) Fields]\n\n");
  MemIO_WriteProc( (thandle_t) outIOBuf, text, strlen(text)-1 );  
  
  sprintf(text, "Name: %s\n\n", meta->Name);
  MemIO_WriteProc( (thandle_t) outIOBuf, text, strlen(text)-1 );
  
  sprintf(text, "AutoScale: %d\n\n", meta->AutoScale);
  MemIO_WriteProc( (thandle_t) outIOBuf, text, strlen(text)-1 );

  sprintf(text, "MinScale: %d\n\n", meta->MinScale);
  MemIO_WriteProc( (thandle_t) outIOBuf, text, strlen(text)-1 );

  sprintf(text, "MaxScale: %d\n\n", meta->MaxScale);
  MemIO_WriteProc( (thandle_t) outIOBuf, text, strlen(text)-1 );

  sprintf(text, "SpatialCalibration: %d\n\n", meta->SpatialCalibration);
  MemIO_WriteProc( (thandle_t) outIOBuf, text, strlen(text)-1 );

  sprintf(text, "XCalibration: %d/%d\n\n", meta->XCalibration[0], meta->XCalibration[1]);
  MemIO_WriteProc( (thandle_t) outIOBuf, text, strlen(text)-1 );

  sprintf(text, "YCalibration: %d/%d\n\n", meta->YCalibration[0], meta->YCalibration[1]);
  MemIO_WriteProc( (thandle_t) outIOBuf, text, strlen(text)-1 );

  sprintf(text, "CalibrationUnits: %s\n\n", meta->CalibrationUnits);
  MemIO_WriteProc( (thandle_t) outIOBuf, text, strlen(text)-1 );

  sprintf(text, "ThreshState: %d\n\n", meta->ThreshState);
  MemIO_WriteProc( (thandle_t) outIOBuf, text, strlen(text)-1 );

  sprintf(text, "ThreshStateRed: %d\n\n", meta->ThreshStateRed);
  MemIO_WriteProc( (thandle_t) outIOBuf, text, strlen(text)-1 );

  sprintf(text, "ThreshStateGreen: %d\n\n", meta->ThreshStateGreen);
  MemIO_WriteProc( (thandle_t) outIOBuf, text, strlen(text)-1 );

  sprintf(text, "ThreshStateBlue: %d\n\n", meta->ThreshStateBlue);
  MemIO_WriteProc( (thandle_t) outIOBuf, text, strlen(text)-1 );

  sprintf(text, "ThreshStateLo: %d\n\n", meta->ThreshStateLo);
  MemIO_WriteProc( (thandle_t) outIOBuf, text, strlen(text)-1 );

  sprintf(text, "ThreshStateHi: %d\n\n", meta->ThreshStateHi);
  MemIO_WriteProc( (thandle_t) outIOBuf, text, strlen(text)-1 );

  sprintf(text, "Zoom: %d\n\n", meta->Zoom);
  MemIO_WriteProc( (thandle_t) outIOBuf, text, strlen(text)-1 );

  // new fields
  int i;
  for (i=0; i<meta->N; i++)
  {
    unsigned short y;
    unsigned char m, d;
    unsigned char h, mi, s;

    JulianToYMD(meta->creationDate[i], y, m, d);
    MiliMidnightToHMS(meta->creationTime[i], h, mi, s);

    sprintf(text, "Page %3.3d: %.4d-%.2d-%.2d %.2d:%.2d:%.2d\n\n", i , y, m, d, h, mi, s );

    MemIO_WriteProc( (thandle_t) outIOBuf, text, strlen(text)-1 ); 
  }

}










