/*****************************************************************************
  TIFF support 
  Copyright (c) 2004 by Dmitry V. Fedorov <www.dimin.net> <dima@dimin.net>

  IMPLEMENTATION
  
  Programmer: Dima V. Fedorov <mailto:dima@dimin.net> <http://www.dimin.net/>

  History:
    03/29/2004 22:23 - First creation
    01/23/2007 20:42 - fixes in warning reporting
        
  Ver : 4
*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "dim_xtiffio.h"
#include "dim_tiff_format.h"

void getCurrentPageInfo(TDimTiffParams *tiffParams);
void getImageInfo(TDimTiffParams *tiffParams);
TDimImageInfo initTDimImageInfo();
void* DimMalloc(DIM_ULONG size);
void* DimFree(void *p);

#include "dim_tiff_format_io.cpp"

//****************************************************************************
// STATIC FUNCTIONS THAT MIGHT BE PROVIDED BY HOST AND CALLING STUBS FOR THEM
//****************************************************************************

static TDimProgressProc  hostProgressProc  = NULL;
static TDimErrorProc     hostErrorProc     = NULL;
static TDimTestAbortProc hostTestAbortProc = NULL;
static TDimMallocProc    hostMallocProc    = NULL;
static TDimFreeProc      hostFreeProc      = NULL;

static void localTiffProgressProc(long done, long total, char *descr)
{
  if (hostProgressProc != NULL) hostProgressProc( done, total, descr );
}

static void localTiffErrorProc(int val, char *descr)
{
  if (hostErrorProc != NULL) hostErrorProc( val, descr );
}

static bool localTiffTestAbortProc ( void )
{
  if (hostTestAbortProc != NULL) 
  {
    if (hostTestAbortProc() == 1) return true; else return false;
  }
  else
    return FALSE;
}

static void* localTiffMallocProc (DIM_ULONG size)
{
  if (hostMallocProc != NULL) 
    return hostMallocProc( size );
  else
  {
    //void *p = (void *) new char[size];
    void *p = (void *) _TIFFmalloc(size);  
    return p;
  }
}

static void* localTiffFreeProc (void *p)
{
  if (hostFreeProc != NULL) 
    return hostFreeProc( p );
  else
  {
    //unsigned char *pu = (unsigned char*) p;
    //if (p != NULL) delete pu;
	  
	if (p != NULL) _TIFFfree( p );
    return NULL;
  }
}

static void setLocalTiffFunctions ( TDimFormatHandle *fmtHndl )
{
  hostProgressProc  = fmtHndl->showProgressProc; 
  hostErrorProc     = fmtHndl->showErrorProc; 
  hostTestAbortProc = fmtHndl->testAbortProc; 
  hostMallocProc    = fmtHndl->mallocProc; 
  hostFreeProc      = fmtHndl->freeProc; 
}

static void resetLocalTiffFunctions (  )
{
  hostProgressProc  = NULL; 
  hostErrorProc     = NULL; 
  hostTestAbortProc = NULL; 
  hostMallocProc    = NULL; 
  hostFreeProc      = NULL; 
}

void* DimMalloc(DIM_ULONG size)
{
  return localTiffMallocProc ( size );
}

void* DimFree(void *p)
{
  return localTiffFreeProc (p);
}


//****************************************************************************
// CALLBACKS
//****************************************************************************

static tsize_t tiff_read(thandle_t handle, tdata_t data, tsize_t size) 
{
  TDimFormatHandle *fmtHndl = (TDimFormatHandle *) handle;
  //if ( fmtHndl->io_mode != DIM_IO_READ ) return 0;
  return (tsize_t) dimRead( fmtHndl, data, 1, size );
}

static tsize_t tiff_write(thandle_t handle,tdata_t data,tsize_t size) 
{
  TDimFormatHandle *fmtHndl = (TDimFormatHandle *) handle;
  if ( fmtHndl->io_mode != DIM_IO_WRITE ) return 0;
  tsize_t res = (tsize_t) dimWrite( fmtHndl, data, 1, size );
  return res;
  //return size;
}

static toff_t tiff_seek(thandle_t handle, toff_t offset, int whence) 
{
  TDimFormatHandle *fmtHndl = (TDimFormatHandle *) handle;
  return (tsize_t) dimSeek( fmtHndl, offset, whence );
}

static int tiff_close(thandle_t handle) {
  TDimFormatHandle *fmtHndl = (TDimFormatHandle *) handle;
  dimFlush( fmtHndl );
  return dimClose( fmtHndl );
}

static toff_t tiff_size(thandle_t handle) {
  TDimFormatHandle *fmtHndl = (TDimFormatHandle *) handle;
  return dimSize( fmtHndl );
}

static int tiff_mmap(thandle_t handle,tdata_t* data,toff_t* size) {
  handle=handle; data=data; size=size;
  return 1;
}

static void tiff_unmap(thandle_t handle, tdata_t data, toff_t size) {
  handle=handle; data=data; size=size;
}



//****************************************************************************
//
// FORMAT DEMANDED FUNTIONS
//
//****************************************************************************

//----------------------------------------------------------------------------
// UTILITARY FUNCTIONS
//----------------------------------------------------------------------------

unsigned int tiffGetNumberOfPages( TDimTiffParams *tiffpar )
{
  unsigned int i=0;
  TIFF *tif = tiffpar->dimTiff;
  if (tif == NULL) return i;
  
  // if STK then get number of pages in special way
  if (tiffpar->subType == tstStk) {
    return stkGetNumPlanes( tif );
  }

  TIFFSetDirectory(tif, 0);
  while (TIFFLastDirectory(tif) == 0) {
    if ( TIFFReadDirectory(tif) == 0) break;
    i++;
  }
  i++;

  //if (tiffpar->subType == tstPsia) i =* 2;

  return i;
}

void tiffReadResolution( TIFF *tif, DIM_UINT &units, double &xRes, double &yRes)
{
  if (tif == NULL) return;
  
  float xresolution = 0;
  float yresolution = 0;
  short resolutionunit = 0;

  units = 0;
  xRes = 0;
  yRes = 0;

  if ( ( TIFFGetField(tif, TIFFTAG_RESOLUTIONUNIT , &resolutionunit) )
       &&
       ( TIFFGetField(tif, TIFFTAG_XRESOLUTION , &xresolution) )
       &&
       ( TIFFGetField(tif, TIFFTAG_YRESOLUTION , &yresolution) )
     )
  {
    units = resolutionunit;
    xRes = xresolution;
    yRes = yresolution;
  }

  // here we need to read specific info here to define resolution correctly


}

DIM_UINT getTiffMode( TIFF *tif)
{
  if (tif == NULL) return DIM_GRAYSCALE;

  uint16 photometric = PHOTOMETRIC_MINISWHITE;
  uint16 samplesperpixel = 1;
  TIFFGetField(tif, TIFFTAG_PHOTOMETRIC, &photometric);

  if (photometric == PHOTOMETRIC_RGB) return DIM_RGB;
  if (photometric == PHOTOMETRIC_PALETTE) return DIM_INDEXED;

  TIFFGetField(tif, TIFFTAG_SAMPLESPERPIXEL, &samplesperpixel);
  if (samplesperpixel > 1) return DIM_MULTI;
    
  return DIM_GRAYSCALE;
}

void getCurrentPageInfo(TDimTiffParams *tiffParams)
{
  if (tiffParams == NULL) return;
  TIFF *tif = tiffParams->dimTiff;
  TDimImageInfo *info = &tiffParams->info;
  
  if (tif == NULL) return;

  uint32 height = 0; 
  uint32 width = 0; 
  uint16 bitspersample = 1;
  uint16 samplesperpixel = 1;

  TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &width);
  TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &height);
  TIFFGetField(tif, TIFFTAG_SAMPLESPERPIXEL, &samplesperpixel);
  TIFFGetField(tif, TIFFTAG_BITSPERSAMPLE, &bitspersample);

  info->width = width;
  info->height = height;

  info->samples = samplesperpixel;
  info->depth = bitspersample;
  if (bitspersample != 16) 
    info->pixelType = DIM_TAG_BYTE;
  else
    info->pixelType = DIM_TAG_SHORT;


  if( !TIFFIsTiled(tif) )
  {
    info->tileWidth = 0;
    info->tileHeight = 0; 
  }
  else
  {
    uint32 columns, rows;
    TIFFGetField(tif, TIFFTAG_TILEWIDTH,  &columns);
    TIFFGetField(tif, TIFFTAG_TILELENGTH, &rows);
    info->tileWidth = columns;
    info->tileHeight = rows; 
  }
    
  info->transparentIndex = 0;
  info->transparencyMatting = 0;

  if (tiffParams->subType != tstFluoview)
    init_image_palette( tif, info );
  else
    fluoviewInitPalette( tiffParams, info);

  info->imageMode = getTiffMode( tif );
  tiffReadResolution( tif, info->resUnits, info->xRes, info->yRes);

  if ( tiffParams->subType == tstFluoview ) 
  {
    TDimFluoviewInfo *fvi = &tiffParams->fluoviewInfo;    
    info->resUnits = DIM_RES_um;
    info->xRes = fvi->xR;
    info->yRes = fvi->yR;
  }

  if ( tiffParams->subType == tstPsia ) 
  {
    psiaInfoHeader *meta = &tiffParams->psiaInfo;
    info->resUnits = DIM_RES_um;
    info->xRes = meta->dfXScanSize / meta->nWidth;
    info->yRes = meta->dfYScanSize / meta->nHeight;
  }

  if ( tiffParams->subType == tstCzLsm ) 
  {
    TDimLsmInfo *lsm = &tiffParams->lsmInfo;  
    info->resUnits = DIM_RES_um;
    info->xRes = lsm->res[0] * 1000000;
    info->yRes = lsm->res[1] * 1000000;
  }

}

void getImageInfo(TDimTiffParams *tiffParams)
{
  if (tiffParams == NULL) return;
  TIFF *tif = tiffParams->dimTiff;
  TDimImageInfo *info = &tiffParams->info;
  if (tif == NULL) return;

  info->ver = sizeof(TDimImageInfo);
  
  // read to which tiff sub type image pertence
  tiffParams->subType = tstGeneric;

  if (stkIsTiffValid( tiffParams ) == TRUE ) {
    tiffParams->subType = tstStk;
    stkGetInfo( tiffParams );
  }

  if (psiaIsTiffValid( tiffParams ) == TRUE ) {
    tiffParams->subType = tstPsia;
    psiaGetInfo ( tiffParams );
  }

  info->number_pages = tiffGetNumberOfPages( tiffParams );

  // we need to get fluoview parameters knowing the number of pages in tiff
  if (fluoviewIsTiffValid( tiffParams ) == TRUE ) {
    tiffParams->subType = tstFluoview;
    fluoviewGetInfo ( tiffParams );
  }

  // lsm has thumbnails for each image, discard those
  if (lsmIsTiffValid( tiffParams ) == TRUE ) {
    tiffParams->subType = tstCzLsm;
    lsmGetInfo ( tiffParams );
  }

  // set dimensions
  if (info->number_pages > 1) {

    if (tiffParams->subType == tstStk) {
      info->number_dims = 4;
      info->dimensions[3].dim = DIM_DIM_Z;
      info->number_t = info->number_pages;
    }
  }

  getCurrentPageInfo( tiffParams );
}

//----------------------------------------------------------------------------
// PARAMETERS, INITS
//----------------------------------------------------------------------------

DIM_INT dimTiffValidateFormatProc (DIM_MAGIC_STREAM *magic, DIM_UINT length)
{
  if (length < 4) return -1;

  if (memcmp(magic,"\115\115\000\052",4) == 0) return 0;
  if (memcmp(magic,"\111\111\052\000",4) == 0) return 0;

  return -1;
}

TDimFormatHandle dimTiffAquireFormatProc( void )
{
  TDimFormatHandle fp = initTDimFormatHandle();
  
  return fp;
}


void dimTiffReleaseFormatProc (TDimFormatHandle *fmtHndl)
{
  if (fmtHndl == NULL) return;
  dimTiffCloseImageProc ( fmtHndl );  
  resetLocalTiffFunctions( );
}

//----------------------------------------------------------------------------
// OPEN/CLOSE
//----------------------------------------------------------------------------
void dimTiffSetWriteParameters (TDimFormatHandle *fmtHndl) {
  if (fmtHndl == NULL) return;
  fmtHndl->compression = COMPRESSION_LZW;

  if (!fmtHndl->options) return;
  xstring str = fmtHndl->options;
  std::vector<xstring> options = str.split( " " );
  if (options.size() < 1) return;
  
  int i = -1;
  while (i<(int)options.size()-1) {
    i++;

    if ( options[i]=="compression" && options.size()-i>0 ) {
      i++;
      if (options[i] == "none") fmtHndl->compression = COMPRESSION_NONE;
      if (options[i] == "fax") fmtHndl->compression = COMPRESSION_CCITTFAX4;
      if (options[i] == "lzw") fmtHndl->compression = COMPRESSION_LZW;
      if (options[i] == "packbits") fmtHndl->compression = COMPRESSION_PACKBITS;
      continue;
    }
  } // while
}

void dimTiffCloseImageProc (TDimFormatHandle *fmtHndl)
{
  if (fmtHndl == NULL) return;
  if (fmtHndl->internalParams == NULL) return;

  TDimTiffParams *tiffpar = (TDimTiffParams *) fmtHndl->internalParams;

  if (fmtHndl->io_mode != DIM_IO_WRITE)
  {
    if (tiffpar != NULL) clearTiffIFDs( &tiffpar->ifds );
    clearMetaTags( &fmtHndl->metaData );
  }

  if ( (tiffpar != NULL) && (tiffpar->dimTiff != NULL) ) {
    XTIFFClose( tiffpar->dimTiff );
    tiffpar->dimTiff = NULL;
  }

  // here clear STK struct
  if ( (tiffpar != NULL) && (tiffpar->subType == tstStk) )
    stkClearInfo ( tiffpar );

  if (fmtHndl->internalParams != NULL) {
    delete tiffpar;
    fmtHndl->internalParams = NULL;
  }
}

DIM_UINT dimTiffOpenImageProc  (TDimFormatHandle *fmtHndl, DIM_ImageIOModes io_mode)
{
  if (fmtHndl == NULL) return 1;
  setLocalTiffFunctions( fmtHndl );

  dimTiffCloseImageProc( fmtHndl );
  TDimTiffParams *tiffpar = new TDimTiffParams;
  fmtHndl->internalParams = tiffpar;  

  tiffpar->info = initTDimImageInfo();
  tiffpar->dimTiff = NULL;
  tiffpar->subType = tstGeneric;
  tiffpar->ifds = initTDimTiffIFDs();
  stkInitInfo( &tiffpar->stkInfo );

  TIFFSetWarningHandler(0);
  TIFFSetErrorHandler(0);

  if (io_mode == DIM_IO_WRITE) {

    if ( isCustomWriting ( fmtHndl ) != TRUE )
      tiffpar->dimTiff = XTIFFOpen(fmtHndl->fileName, "w");
    else 
      tiffpar->dimTiff = XTIFFClientOpen( fmtHndl->fileName, "wm",
        (thandle_t) fmtHndl, tiff_read, tiff_write, tiff_seek, tiff_close, tiff_size, tiff_mmap, tiff_unmap );
    dimTiffSetWriteParameters (fmtHndl);  
  } else { // if reading

    if ( isCustomReading ( fmtHndl ) != TRUE )
      tiffpar->dimTiff = XTIFFOpen(fmtHndl->fileName, "r");
    else 
      tiffpar->dimTiff = XTIFFClientOpen( fmtHndl->fileName, "rm",
        (thandle_t) fmtHndl, tiff_read, tiff_write, tiff_seek, tiff_close, tiff_size, tiff_mmap, tiff_unmap );

    if (tiffpar->dimTiff != NULL) {
      tiffpar->ifds = readAllTiffIFDs( tiffpar->dimTiff );
      getImageInfo(tiffpar);
      fmtHndl->subFormat = tiffpar->subType;
    }
  }

  if (tiffpar->dimTiff == NULL) return 1;

  return 0;
}

DIM_UINT dimTiffFOpenImageProc (TDimFormatHandle *fmtHndl, char* fileName, DIM_ImageIOModes io_mode)
{
  fmtHndl->fileName = fileName;
  return dimTiffOpenImageProc(fmtHndl, io_mode);
}

DIM_UINT dimTiffIOpenImageProc (TDimFormatHandle *fmtHndl, char* fileName, 
                                         DIM_IMAGE_CLASS *image, DIM_ImageIOModes io_mode)
{
  fmtHndl->fileName = fileName;
  fmtHndl->image    = image;
  return dimTiffOpenImageProc(fmtHndl, io_mode);
}


//----------------------------------------------------------------------------
// INFO for OPEN image
//----------------------------------------------------------------------------

DIM_UINT dimTiffGetNumPagesProc ( TDimFormatHandle *fmtHndl )
{
  if (fmtHndl == NULL) return 0;
  if (fmtHndl->internalParams == NULL) return 0;
  TDimTiffParams *tiffpar = (TDimTiffParams *) fmtHndl->internalParams;

  if (tiffpar->dimTiff == NULL) return 0;

  return tiffpar->info.number_pages;
}


TDimImageInfo dimTiffGetImageInfoProc ( TDimFormatHandle *fmtHndl, DIM_UINT page_num )
{
  TDimImageInfo ii = initTDimImageInfo();

  if (fmtHndl == NULL) return ii;
  if (fmtHndl->internalParams == NULL) return ii;
  TDimTiffParams *tiffpar = (TDimTiffParams *) fmtHndl->internalParams;
  TIFF *tif = tiffpar->dimTiff;
  if (tif == NULL) return ii;

  fmtHndl->pageNumber = page_num;
  fmtHndl->subFormat = tiffpar->subType;

  unsigned int currentDir = TIFFCurrentDirectory(tif);

  // now must read correct page and set image parameters
  if (currentDir != fmtHndl->pageNumber)
  if (tiffpar->subType != tstStk)
  {
    TIFFSetDirectory(tif, fmtHndl->pageNumber);
    getCurrentPageInfo( tiffpar );
  }

  ii = tiffpar->info; 

  return ii;
}

//----------------------------------------------------------------------------
// METADATA
//----------------------------------------------------------------------------

// libTIFF CANNOT ADD TAGS INTO ANY GIVEN IMAGE
DIM_UINT dimTiffAddMetaDataProc (TDimFormatHandle *fmtHndl)
{
  fmtHndl=fmtHndl;
  return 1;
}


DIM_UINT dimTiffReadMetaDataProc (TDimFormatHandle *fmtHndl, DIM_UINT page, int group, int tag, int type)
{
  if (fmtHndl == NULL) return 1;
  if (fmtHndl->internalParams == NULL) return 1;
  TDimTiffParams *tiffpar = (TDimTiffParams *) fmtHndl->internalParams;
  fmtHndl->pageNumber = page;
  
  return read_tiff_metadata (fmtHndl, tiffpar, group, tag, type);
}

char* dimTiffReadMetaDataAsTextProc ( TDimFormatHandle *fmtHndl )
{
  if (fmtHndl == NULL) return NULL;
  if (fmtHndl->internalParams == NULL) return NULL;
  TDimTiffParams *tiffpar = (TDimTiffParams *) fmtHndl->internalParams;
  
  return read_text_tiff_metadata ( fmtHndl, tiffpar );
}


//----------------------------------------------------------------------------
// READ/WRITE
//----------------------------------------------------------------------------

DIM_UINT dimTiffReadImageProc  ( TDimFormatHandle *fmtHndl, DIM_UINT page )
{
  if (fmtHndl == NULL) return 1;
  if (fmtHndl->internalParams == NULL) return 1;
  TDimTiffParams *tiffpar = (TDimTiffParams *) fmtHndl->internalParams;

  if (tiffpar->dimTiff == NULL) return 1;
  fmtHndl->pageNumber = page;
  
  return read_tiff_image(fmtHndl, tiffpar);
}

DIM_UINT dimTiffWriteImageProc ( TDimFormatHandle *fmtHndl )
{
  if (fmtHndl == NULL) return 1;
  if (fmtHndl->internalParams == NULL) return 1;
  TDimTiffParams *tiffpar = (TDimTiffParams *) fmtHndl->internalParams;

  if (tiffpar->dimTiff == NULL) return 1;

  return write_tiff_image(fmtHndl, tiffpar);

  return 0;
}

// at the moment w and h make no effect,the image retreived is the same size as original
DIM_UINT dimTiffReadImagePreviewProc (TDimFormatHandle *fmtHndl, DIM_UINT w, DIM_UINT h)
{
  TDimImageBitmap bmp8, *bmp;
  initImagePlanes( &bmp8 );

  w=w; h=h;
  
  if ( dimTiffReadImageProc( fmtHndl, 0 ) != 0) return 1;
  bmp = fmtHndl->image;
  
  if (bmp->i.samples > 3) bmp->i.samples = 3;

  if (bmp->i.depth == 16) {
    allocImg(&bmp8, bmp->i.width, bmp->i.height, bmp->i.samples, 8);
    normalizeImg(bmp, &bmp8);
    deleteImg( bmp );

    bmp->bits[0] = bmp8.bits[0]; 
    bmp->bits[1] = bmp8.bits[1];    
    bmp->bits[2] = bmp8.bits[2]; 
  }

  bmp->i.depth = 8;
  
  if (bmp->i.samples == 1) {
    bmp->bits[1] = bmp->bits[0];    
    bmp->bits[2] = bmp->bits[0]; 
  }

  if (bmp->i.samples == 2) {
    bmp->bits[2] = bmp->bits[0]; 
  }
  bmp->i.samples = 3;

  return 0;
}


//****************************************************************************
//
// EXPORTED FUNCTION
//
//****************************************************************************

TDimFormatItem dimTiffItems[5] = {
  {
    "TIFF",            // short name, no spaces
    "Tagged Image File Format (GeoTIFF)", // Long format name
    "tif|tiff|fax|geotiff",        // pipe "|" separated supported extension list
    1, //canRead;      // 0 - NO, 1 - YES
    1, //canWrite;     // 0 - NO, 1 - YES
    1, //canReadMeta;  // 0 - NO, 1 - YES
    0, //canWriteMeta; // 0 - NO, 1 - YES
    1, //canWriteMultiPage;   // 0 - NO, 1 - YES
    //TDivFormatConstrains constrains ( w, h, pages, minsampl, maxsampl, minbitsampl, maxbitsampl, noLut )
    { 0, 0, 0, 0, 0, 0, 0, 0 }
  },
  {
    "STK",            // short name, no spaces
    "Metamorph Stack File Format", // Long format name
    "stk",        // pipe "|" separated supported extension list
    1, //canRead;      // 0 - NO, 1 - YES
    0, //canWrite;     // 0 - NO, 1 - YES
    1, //canReadMeta;  // 0 - NO, 1 - YES
    0, //canWriteMeta; // 0 - NO, 1 - YES
    0, //canWriteMultiPage;   // 0 - NO, 1 - YES
    //TDivFormatConstrains constrains ( w, h, pages, minsampl, maxsampl, minbitsampl, maxbitsampl, noLut )    
    { 0, 0, 0, 0, 0, 0, 0, 0 }
  },
  {
    "PSIA",            // short name, no spaces
    "AFM PSIA TIFF File Format", // Long format name
    "tif|tiff",        // pipe "|" separated supported extension list
    1, //canRead;      // 0 - NO, 1 - YES
    0, //canWrite;     // 0 - NO, 1 - YES
    1, //canReadMeta;  // 0 - NO, 1 - YES
    0, //canWriteMeta; // 0 - NO, 1 - YES
    0, //canWriteMultiPage;   // 0 - NO, 1 - YES
    //TDivFormatConstrains constrains ( w, h, pages, minsampl, maxsampl, minbitsampl, maxbitsampl, noLut )    
    { 0, 0, 0, 0, 0, 0, 0, 0 }
  },
  {
    "FLUOVIEW",            // short name, no spaces
    "Fluoview TIFF File Format", // Long format name
    "tif|tiff",        // pipe "|" separated supported extension list
    1, //canRead;      // 0 - NO, 1 - YES
    0, //canWrite;     // 0 - NO, 1 - YES
    1, //canReadMeta;  // 0 - NO, 1 - YES
    0, //canWriteMeta; // 0 - NO, 1 - YES
    0, //canWriteMultiPage;   // 0 - NO, 1 - YES
    //TDivFormatConstrains constrains ( w, h, pages, minsampl, maxsampl, minbitsampl, maxbitsampl, noLut )    
    { 0, 0, 0, 0, 0, 0, 0, 0 }
  },
  {
    "LSM",            // short name, no spaces
    "Carl Zeiss LSM5 File Format", // Long format name
    "lsm",        // pipe "|" separated supported extension list
    1, //canRead;      // 0 - NO, 1 - YES
    0, //canWrite;     // 0 - NO, 1 - YES
    1, //canReadMeta;  // 0 - NO, 1 - YES
    0, //canWriteMeta; // 0 - NO, 1 - YES
    0, //canWriteMultiPage;   // 0 - NO, 1 - YES
    //TDivFormatConstrains constrains ( w, h, pages, minsampl, maxsampl, minbitsampl, maxbitsampl, noLut )    
    { 0, 0, 0, 0, 0, 0, 0, 0 }
  }
};

TDimFormatHeader dimTiffHeader = {

  sizeof(TDimFormatHeader),
  "1.0.2",
  "DIMIN TIFF CODEC",
  "Tagged Image File Format CODEC",
  
  4,                      // 0 or more, specify number of bytes needed to identify the file
  {1, 5, dimTiffItems},   //dimTiffSupported,
  
  dimTiffValidateFormatProc,
  // begin
  dimTiffAquireFormatProc, //TDimAquireFormatProc
  // end
  dimTiffReleaseFormatProc, //TDimReleaseFormatProc
  
  // params
  NULL, //TDimAquireIntParamsProc
  NULL, //TDimLoadFormatParamsProc
  NULL, //TDimStoreFormatParamsProc

  // image begin
  dimTiffOpenImageProc, //TDimOpenImageProc
  dimTiffCloseImageProc, //TDimCloseImageProc 

  // info
  dimTiffGetNumPagesProc, //TDimGetNumPagesProc
  dimTiffGetImageInfoProc, //TDimGetImageInfoProc


  // read/write
  dimTiffReadImageProc, //TDimReadImageProc 
  dimTiffWriteImageProc, //TDimWriteImageProc
  NULL, //TDimReadImageTileProc
  NULL, //TDimWriteImageTileProc
  NULL, //TDimReadImageLineProc
  NULL, //TDimWriteImageLineProc
  NULL, //TDimReadImageThumbProc
  NULL, //TDimWriteImageThumbProc
  dimTiffReadImagePreviewProc, //TDimReadImagePreviewProc
  
  // meta data
  dimTiffReadMetaDataProc, //TDimReadMetaDataProc
  dimTiffAddMetaDataProc,  //TDimAddMetaDataProc
  dimTiffReadMetaDataAsTextProc, //TDimReadMetaDataAsTextProc

  NULL,
  NULL,
  NULL,
  ""

};

extern "C" {

TDimFormatHeader* dimTiffGetFormatHeader(void)
{
  return &dimTiffHeader;
}

} // extern C





