/*******************************************************************************

  histogram and lutf
  
  Author: Dima Fedorov Levit <dimin@dimin.net> <http://www.dimin.net/>

  History:
    10/20/2006 20:21 - First creation
    2007-06-26 12:18 - added lut class

  ver: 2
        
*******************************************************************************/

#include "dim_histogram.h"
#include "xtypes.h"

#include "cmath"

//******************************************************************************
// DimHistogram
//******************************************************************************

DimHistogram::DimHistogram( int bpp ) {
  init( bpp ); 
}

DimHistogram::DimHistogram( int bpp, void *data, unsigned int num_data_points ) {
  init( bpp );
  addData( data, num_data_points );
}

DimHistogram::~DimHistogram() {

}

void DimHistogram::init( int bpp ) {
  data_bpp = bpp;
  if (bpp == 0)
    hist.resize( 0 );  
  else
    hist.resize( (unsigned int) pow(2.0, abs(bpp)), 0 );
}

void DimHistogram::clear( ) {
  hist = std::vector<unsigned int>(hist.size(), 0); 
}

void DimHistogram::newData( int bpp, void *data, unsigned int num_data_points ) {
  init( bpp );
  clear( );
  addData( data, num_data_points );
}

template <typename T>
void DimHistogram::add_from_data( T *data, unsigned int num_data_points ) {
  if (!data) return;
  if (num_data_points==0) return;

  T *p = (T *) data;  
  for (unsigned int i=0; i<num_data_points; ++i) {
    ++hist[*p];
    ++p;
  }
}

void DimHistogram::addData( void *data, unsigned int num_data_points ) {

  if (data_bpp == 8)
    add_from_data<unsigned char>( (unsigned char*)data, num_data_points );

  if (data_bpp == -8)
    add_from_data<char>( (char*)data, num_data_points );

  if (data_bpp == 16)
    add_from_data<unsigned short>( (unsigned short*)data, num_data_points );

  if (data_bpp == -16)
    add_from_data<short>( (short*)data, num_data_points );
}

int DimHistogram::last_pos() const {
  for (unsigned int i=hist.size()-1; i>=0; --i)
    if (hist[i] != 0)
      return i;
  return 0;
}

int DimHistogram::first_pos() const {
  for (unsigned int i=0; i<hist.size(); ++i)
    if (hist[i] != 0)
      return i;
  return 0;
}

int DimHistogram::max_pos() const {
  int pos = 0;
  unsigned int val = hist[0];
  for (unsigned int i=0; i<hist.size(); ++i)
    if (hist[i] > val) {
      val = hist[i];
      pos = i;
    }
  return pos;
}

int DimHistogram::min_pos() const {
  int pos = 0;
  unsigned int val = hist[0];
  for (unsigned int i=0; i<hist.size(); ++i)
    if (hist[i] < val) {
      val = hist[i];
      pos = i;
    }
  return pos;
}

int DimHistogram::num_unique() const {
  int unique = 0;
  for (unsigned int i=0; i<hist.size(); ++i) 
    if (hist[i] != 0) ++unique;
  return unique;
}

unsigned int DimHistogram::cumsum( unsigned int pos ) const {
  if ( pos >= hist.size() ) pos = hist.size()-1;
  unsigned int sum=0;
  for (unsigned int i=0; i<=pos; ++i) 
    sum += hist[i];
  return sum;
}

unsigned int DimHistogram::get_value( unsigned int pos ) const {
  if ( pos < hist.size() )
    return hist[pos];
  else
    return 0;
}

void DimHistogram::set_value( unsigned int pos, unsigned int val ) {
  if ( pos < hist.size() )
    hist[pos] = val;
}

void DimHistogram::append_value( unsigned int pos, unsigned int val ) {
  if ( pos < hist.size() )
    hist[pos] += val;
}

//******************************************************************************
// DimLut
//******************************************************************************

DimLut::DimLut( ) {
  in_depth=0; out_depth=0;
  lut.resize( 0 ); 
  this->generator = linear_full_range_generator;
}

DimLut::DimLut( const DimHistogram &in, const DimHistogram &out ) {
  init( in, out );
}

DimLut::DimLut( const DimHistogram &in, const DimHistogram &out, LutType type ) {
  init( in, out, type );
}

DimLut::DimLut( const DimHistogram &in, const DimHistogram &out, DimLutGenerator custom_generator ) {
  init( in, out, custom_generator );
}

DimLut::~DimLut() {

}

void DimLut::init( const DimHistogram &in, const DimHistogram &out ) {
  lut.resize( in.size() );
  clear( );
  if (generator) generator( in, lut, out.size(), NULL );
  in_depth  = abs(in.dataBpp());
  out_depth = abs(out.dataBpp());
}

void DimLut::init( const DimHistogram &in, const DimHistogram &out, LutType type ) {
  if (type == ltLinearFullRange)     generator = linear_full_range_generator;
  if (type == ltLinearDataRange)     generator = linear_data_range_generator;
  if (type == ltLinearDataTolerance) generator = linear_data_tolerance_generator;
  if (type == ltEqualize)            generator = equalized_generator;
  init( in, out );
}

void DimLut::init( const DimHistogram &in, const DimHistogram &out, DimLutGenerator custom_generator ) {
  generator = custom_generator;
  init( in, out );
}

void DimLut::clear( ) {
  lut = std::vector<unsigned int>(lut.size(), 0); 
}

template <typename Tl>
void DimLut::set_lut( const std::vector<Tl> &new_lut ) {
  lut.assign( new_lut.begin(), new_lut.end() );
}
template void DimLut::set_lut<unsigned char>( const std::vector<unsigned char> &new_lut );
template void DimLut::set_lut<unsigned int>( const std::vector<unsigned int> &new_lut );

unsigned int DimLut::get_value( unsigned int pos ) const {
  if ( pos < lut.size() )
    return lut[pos];
  else
    return 0;
}

void DimLut::set_value( unsigned int pos, unsigned int val ) {
  if ( pos < lut.size() )
    lut[pos] = val;
}

template <typename Ti, typename To>
void DimLut::apply_lut( const Ti *ibuf, To *obuf, unsigned int num_data_points ) const {
  if (!ibuf || !obuf) return;

  for (unsigned int i=0; i<num_data_points; ++i)
    obuf[i] = (To) lut[ ibuf[i] ];
}

void DimLut::apply( void *ibuf, const void *obuf, unsigned int num_data_points ) const {
  if (lut.size() <= 0) return;

  if (in_depth == 8 && out_depth == 8)
    apply_lut<unsigned char, unsigned char>( (unsigned char*) ibuf, (unsigned char*) obuf, num_data_points );

  if (in_depth == 16 && out_depth == 8)
    apply_lut<unsigned short, unsigned char>( (unsigned short*) ibuf, (unsigned char*) obuf, num_data_points );

  if (in_depth == 8 && out_depth == 16)
    apply_lut<unsigned char, unsigned short>( (unsigned char*) ibuf, (unsigned short*) obuf, num_data_points );

  if (in_depth == 16 && out_depth == 16)
    apply_lut<unsigned short, unsigned short>( (unsigned short*) ibuf, (unsigned short*) obuf, num_data_points );

}

void DimLut::apply( const DimHistogram &in, DimHistogram &out ) const {
  if (lut.size() <= 0) return;
  out.clear();
  for (unsigned int i=0; i<in.size(); ++i)
    out.append_value( lut[i], in[i] );
}

//******************************************************************************
// Generators
//******************************************************************************

//------------------------
// this is not a generator per si
//------------------------
template <typename Tl>
void linear_range_generator( int b, int e, unsigned int out_phys_range, std::vector<Tl> &lut ) {
  
  //if (lut.size() < out_phys_range) lut.resize(out_phys_range);
  // simple linear mapping for actual range
  double range = e - b;
  if (range < 1) range = out_phys_range;
  for (unsigned int x=0; x<lut.size(); ++x)
    lut[x] = dim::trim<Tl, double> ( (((double)x)-b)*out_phys_range/range, 0, out_phys_range-1);
}
template void linear_range_generator<unsigned char>( int b, int e, unsigned int out_phys_range, std::vector<unsigned char> &lut );
template void linear_range_generator<unsigned int>( int b, int e, unsigned int out_phys_range, std::vector<unsigned int> &lut );
//------------------------


template <typename Tl>
void linear_full_range_generator( const DimHistogram &, std::vector<Tl> &lut, unsigned int out_phys_range, void * ) {
  // simple linear mapping for full range
  for (unsigned int x=0; x<lut.size(); ++x)
    lut[x] = (Tl) ( (((double)x)*(out_phys_range-1.0))/(lut.size()-1.0) );
}
template void linear_full_range_generator<unsigned char>( const DimHistogram &in, std::vector<unsigned char> &lut, unsigned int out_phys_range, void *args );
template void linear_full_range_generator<unsigned int>( const DimHistogram &in, std::vector<unsigned int> &lut, unsigned int out_phys_range, void *args );


template <typename Tl>
void linear_data_range_generator( const DimHistogram &in, std::vector<Tl> &lut, unsigned int out_phys_range, void * ) {
  // simple linear mapping for actual range
  int b = in.first_pos();
  int e = in.last_pos();
  double range = e - b;
  if (range < 1) range = out_phys_range;
  for (unsigned int x=0; x<lut.size(); ++x)
    lut[x] = dim::trim<Tl, double> ( (((double)x)-b)*out_phys_range/range, 0, out_phys_range-1);
}
template void linear_data_range_generator<unsigned char>( const DimHistogram &in, std::vector<unsigned char> &lut, unsigned int out_phys_range, void *args );
template void linear_data_range_generator<unsigned int>( const DimHistogram &in, std::vector<unsigned int> &lut, unsigned int out_phys_range, void *args );

template <typename Tl>
void linear_data_tolerance_generator( const DimHistogram &in, std::vector<Tl> &lut, unsigned int out_phys_range, void * ) {
  
  // simple linear mapping cutting elements with small appearence
  // get 1% threshold
  unsigned int th = (unsigned int) ( in[ in.max_pos() ] * 1.0 / 100.0 );
  int b = 0;
  int e = in.size()-1;
  for (unsigned int x=0; x<in.size(); ++x)
    if ( in[x] > th ) {
      b = x;
      break;
    }
  for (unsigned int x=in.size()-1; x>=0; --x)
    if ( in[x] > th ) {
      e = x;
      break;
    }

  double range = e - b;
  if (range < 1) range = out_phys_range;
  for (unsigned int x=0; x<lut.size(); ++x)
    lut[x] = dim::trim<Tl, double> ( (((double)x)-b)*out_phys_range/range, 0, out_phys_range-1 );
}
template void linear_data_tolerance_generator<unsigned char>( const DimHistogram &in, std::vector<unsigned char> &lut, unsigned int out_phys_range, void *args );
template void linear_data_tolerance_generator<unsigned int>( const DimHistogram &in, std::vector<unsigned int> &lut, unsigned int out_phys_range, void *args );

template <typename Tl>
void equalized_generator( const DimHistogram &in, std::vector<Tl> &lut, unsigned int out_phys_range, void * ) {
  
  // equalize
  std::vector<double> map(lut.size(), 0);
  map[0] = in[0];
  for (unsigned int x=1; x<lut.size(); ++x)
    map[x] = map[x-1] + in[x];

  double div = map[lut.size()-1]-map[0];
  if (div > 0)
  for (unsigned int x=0; x<lut.size(); ++x)
    lut[x] = dim::trim<Tl, double>( (Tl) (out_phys_range-1.0)*((map[x]-map[0]) / div), 0, out_phys_range-1 );
}
template void equalized_generator<unsigned char>( const DimHistogram &in, std::vector<unsigned char> &lut, unsigned int out_phys_range, void *args );
template void equalized_generator<unsigned int>( const DimHistogram &in, std::vector<unsigned int> &lut, unsigned int out_phys_range, void *args );

template <typename Tl>
void linear_custom_range_generator( const DimHistogram &/*in*/, std::vector<Tl> &lut, unsigned int out_phys_range, void *args ) {
  // simple linear mapping for actual range
  int *vals = (int *)args;
  int b = vals[0];
  int e = vals[1];
  double range = e - b;
  if (range < 1) range = out_phys_range;
  for (unsigned int x=0; x<lut.size(); ++x)
    lut[x] = dim::trim<Tl, double> ( (((double)x)-b)*out_phys_range/range, 0, out_phys_range-1);
}
template void linear_custom_range_generator<unsigned char>( const DimHistogram &in, std::vector<unsigned char> &lut, unsigned int out_phys_range, void *args );
template void linear_custom_range_generator<unsigned int>( const DimHistogram &in, std::vector<unsigned int> &lut, unsigned int out_phys_range, void *args );

