/*******************************************************************************

  histogram and lut
  
  Author: Dima Fedorov Levit <dimin@dimin.net> <http://www.dimin.net/>

  History:
    10/20/2006 20:21 - First creation
    2007-06-26 12:18 - added lut class

  ver: 2
        
*******************************************************************************/

#ifndef DIM_HISTOGRAM_H
#define DIM_HISTOGRAM_H

#include <vector>

//******************************************************************************
// DimHistogram
//******************************************************************************

class DimHistogram {
  public:
    DimHistogram( int bpp=0 );
    DimHistogram( int bpp, void *data, unsigned int num_data_points );
    ~DimHistogram();

    void init( int bpp );  // negative bpp is used for signed data
    void clear( );
    void addData( void *data, unsigned int num_data_points );
    void newData( int bpp, void *data, unsigned int num_data_points );
    int dataBpp() const { return data_bpp; }

    int first_pos() const;
    int last_pos() const;
    int max_pos() const;
    int min_pos() const;
    int num_unique() const;
    unsigned int size() const { return hist.size(); }
    const std::vector<unsigned int> &get_hist( ) const { return hist; }

    unsigned int cumsum( unsigned int pos ) const;
    unsigned int get_value( unsigned int pos ) const;
    void set_value( unsigned int pos, unsigned int val );
    void append_value( unsigned int pos, unsigned int val );
    inline unsigned int operator[](unsigned int x) const { return hist[x]; }
    
  protected:
    // 8 = uchar, 16 = ushort, -8 = char, -16 = short and so on
    int data_bpp;
    std::vector<unsigned int> hist;

    template <typename T>
    void add_from_data( T *data, unsigned int num_data_points );

};

//******************************************************************************
// DimLut
//******************************************************************************

class DimLut {
  public:

    enum LutType { 
      ltLinearFullRange=0, 
      ltLinearDataRange=1, 
      ltLinearDataTolerance=2,
      ltEqualize=3
    };

    typedef void (*DimLutGenerator)( const DimHistogram &in, std::vector<unsigned int> &lut, unsigned int out_phys_range, void *args );

    DimLut( );
    DimLut( const DimHistogram &in, const DimHistogram &out );
    DimLut( const DimHistogram &in, const DimHistogram &out, LutType type );
    DimLut( const DimHistogram &in, const DimHistogram &out, DimLutGenerator custom_generator );
    ~DimLut();

    // negative bpp is used for signed data, not implemented yet...
    void init( const DimHistogram &in, const DimHistogram &out );
    void init( const DimHistogram &in, const DimHistogram &out, LutType type );
    void init( const DimHistogram &in, const DimHistogram &out, DimLutGenerator custom_generator );
    void clear( );

    unsigned int size() const { return lut.size(); }
    const std::vector<unsigned int> &get_lut( ) const { return lut; }

    int depthInput()  const { return in_depth; }
    int depthOutput() const { return out_depth; }

    template <typename Tl>
    void         set_lut( const std::vector<Tl> & );
    
    unsigned int get_value( unsigned int pos ) const;
    void         set_value( unsigned int pos, unsigned int val );
    inline unsigned int operator[](unsigned int x) const { return lut[x]; }

    void apply( void *ibuf, const void *obuf, unsigned int num_data_points ) const;
    // generates values of output histogram, given the current lut and in histogram
    void apply( const DimHistogram &in, DimHistogram &out ) const;

  protected:
    std::vector<unsigned int> lut;
    DimLutGenerator generator;
    int in_depth, out_depth;
    void *internal_arguments;

    template <typename Ti, typename To>
    void apply_lut( const Ti *ibuf, To *obuf, unsigned int num_data_points ) const;
};

//******************************************************************************
// Generators - default
//******************************************************************************

// misc
template <typename Tl>
void linear_range_generator( int b, int e, unsigned int out_phys_range, std::vector<Tl> &lut );

template <typename Tl>
void linear_full_range_generator( const DimHistogram &in, std::vector<Tl> &lut, unsigned int out_phys_range, void *args );

template <typename Tl>
void linear_data_range_generator( const DimHistogram &in, std::vector<Tl> &lut, unsigned int out_phys_range, void *args );

template <typename Tl>
void linear_data_tolerance_generator( const DimHistogram &in, std::vector<Tl> &lut, unsigned int out_phys_range, void *args );

template <typename Tl>
void equalized_generator( const DimHistogram &in, std::vector<Tl> &lut, unsigned int out_phys_range, void *args );

// args is a pointer to two int values
template <typename Tl>
void linear_custom_range_generator( const DimHistogram &in, std::vector<Tl> &lut, unsigned int out_phys_range, void *args );

#endif //DIM_HISTOGRAM_H


