/*******************************************************************************

  Defines Image Stack Class
  
  Programmer: Dima Fedorov Levit <dimin@dimin.net> <http://www.dimin.net/>

  History:
    03/23/2004 18:03 - First creation
      
  ver: 1
        
*******************************************************************************/

#ifndef DIM_IMAGE_STACK_H
#define DIM_IMAGE_STACK_H

#include <vector>
#include <map>

#include "tag_map.h"
#include "dim_image.h"
#include "dim_histogram.h"


//------------------------------------------------------------------------------
// DImageStack
//------------------------------------------------------------------------------

class DImageStack {
  public:
    DImageStack();
    ~DImageStack();

    void        free();
    void        clear() { free(); }
    DImageStack deepCopy() const;  

    bool       isReady() const { return !handling_image; } // true if images can be used
    bool       isEmpty() const;

    int        size() const { return images.size(); }
    int        count() const { return size(); }
    int        numberImages() const { return size(); }
    int        numberPlanes() const { return size(); }

    DIM_ULONG    width()     const { if (size()<=0) return 0; else return images[0].width();  }
    DIM_ULONG    height()    const { if (size()<=0) return 0; else return images[0].height(); }
    DIM_ULONG    length()    const { return size(); }
    DIM_UINT     depth()     const { if (size()<=0) return 0; else return images[0].depth();  }
    DIM_UINT     samples()   const { if (size()<=0) return 0; else return images[0].samples(); }

    TDimImage *imageAt( unsigned int position ) const;
    TDimImage *image() const; // image at the currently selected position
    inline TDimImage *operator[](unsigned int p) const { return imageAt(p); }

    void append( const TDimImage &img ) { images.push_back(img); }

    // return true if the level exists
    bool positionPrev();
    bool positionNext();
    bool position0();
    bool positionFirst() { return position0(); }
    bool positionLast();
    bool positionSet( unsigned int l );
    int  positionCurrent() const { return cur_position; }

    const std::map<std::string, std::string> &get_metadata() const { return metadata; }
    std::string get_metadata_tag( const std::string &key, const std::string &def ) const { return metadata.get_value( key, def ); }
    int         get_metadata_tag_int( const std::string &key, const int &def ) const { return metadata.get_value_int( key, def ); }
    double      get_metadata_tag_double( const std::string &key, const double &def ) const { return metadata.get_value_double( key, def ); }

    // I/O

    virtual bool fromFile( const char *fileName );
    bool fromFile( const std::string &fileName ) { return fromFile( fileName.c_str() ); }

    virtual bool toFile( const char *fileName, const char *formatName );
    bool toFile( const std::string &fileName, const std::string &formatName ) {
      return toFile( fileName.c_str(), formatName.c_str() ); }

    // Operations
    void convertToDepth( int depth, DimLut::LutType method = DimLut::ltLinearFullRange, bool planes_independent = false );
    void convertToDepth( const DImageLut & );
    void normalize( int to_bpp = 8, bool planes_independent = false );
    void ensureTypedDepth();

    void resize( int w, int h=0, int d=0, TDimImage::ResizeMethod method = TDimImage::szNearestNeighbor, bool keep_aspect_ratio = false );

    void ROI( DIM_UINT x, DIM_UINT y, DIM_UINT z, DIM_UINT w, DIM_UINT h, DIM_UINT d );

    // only available values now are +90, -90 and 180
    void rotate( double deg );

    void negative();

    TDimImage projectionXZAxis( unsigned int y );
    TDimImage projectionYZAxis( unsigned int x );

  protected:
    std::vector<TDimImage> images;
    DTagMap metadata;

    TDimImage empty_image;
    int  cur_position;
    bool handling_image;

  protected:
    // prohibit copy-constructor
    DImageStack( const DImageStack& ) {}

    void init();

  // callbacks
  public:
    TDimProgressProc progress_proc;
    TDimErrorProc error_proc;
    TDimTestAbortProc test_abort_proc;

  protected:
    void do_progress( long done, long total, char *descr ) {
      if (progress_proc != NULL) progress_proc( done, total, descr );
    }

};


//------------------------------------------------------------------------------
// DStackHistogram
//------------------------------------------------------------------------------

class DStackHistogram: public DImageHistogram {
public:
  DStackHistogram( );
  DStackHistogram( const DImageStack &stack ) { fromImageStack(stack); }
  ~DStackHistogram();

  void fromImageStack( const DImageStack & );
};

#endif //DIM_IMAGE_STACK_H


